# -*- coding: utf-8 -*- #
# Copyright 2025 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Fetch the FloorSetting resource."""

import dataclasses
import json

from googlecloudsdk.api_lib.model_armor import api as model_armor_api
from googlecloudsdk.calliope import base
from googlecloudsdk.calliope import exceptions
from googlecloudsdk.command_lib.model_armor import args as model_armor_args
from googlecloudsdk.command_lib.model_armor import util as model_armor_util
from googlecloudsdk.core import log


@dataclasses.dataclass
class ServiceConfig:
  """Configuration for an integrated service."""

  name: str
  enum: ...
  enforcement_type_arg: str
  enforcement_type_flag_name: str
  cloud_logging_arg: str
  service_floor_setting_arg: str
  service_floor_setting_attr: str
  defaulted_enforcement_flag_name: str
  setting_message: ...
  cli_name: str


def _GetServicesConfig(messages):
  """Returns a list of service configurations."""
  return [
      ServiceConfig(
          name='AI_PLATFORM',
          enum=(
              messages.FloorSetting.IntegratedServicesValueListEntryValuesEnum.AI_PLATFORM
          ),
          enforcement_type_arg='vertex_ai_enforcement_type',
          enforcement_type_flag_name='--vertex-ai-enforcement-type',
          cloud_logging_arg='enable_vertex_ai_cloud_logging',
          service_floor_setting_arg='ai_platform_floor_setting',
          service_floor_setting_attr='aiPlatformFloorSetting',
          defaulted_enforcement_flag_name='vertex_ai_enforcement_defaulted',
          setting_message=messages.AiPlatformFloorSetting,
          cli_name='vertex-ai',
      ),
      ServiceConfig(
          name='GOOGLE_MCP_SERVER',
          enum=(
              messages.FloorSetting.IntegratedServicesValueListEntryValuesEnum.GOOGLE_MCP_SERVER
          ),
          enforcement_type_arg='google_mcp_server_enforcement_type',
          enforcement_type_flag_name='--google-mcp-server-enforcement-type',
          cloud_logging_arg='enable_google_mcp_server_cloud_logging',
          service_floor_setting_attr='googleMcpServerFloorSetting',
          service_floor_setting_arg='google_mcp_server_floor_setting',
          defaulted_enforcement_flag_name=(
              'google_mcp_server_enforcement_defaulted'
          ),
          setting_message=messages.McpServerFloorSetting,
          cli_name='google-mcp-server',
      ),
  ]


@base.DefaultUniverseOnly
@base.ReleaseTracks(
    base.ReleaseTrack.ALPHA, base.ReleaseTrack.BETA, base.ReleaseTrack.GA
)
class Update(base.Command):
  """Update a Model Armor floor setting.

  Updates the Model Armor floor setting with the given name.
  """

  NO_CHANGES_MESSAGE = 'There are no changes to the floor setting for update.'

  @staticmethod
  def Args(parser):
    model_armor_args.AddFullUri(
        parser,
        positional=False,
        required=True,
        help_text='Full uri of the floor setting',
    )
    model_armor_args.AddFloorSettingEnforcement(
        parser,
        positional=False,
        required=False,
    )
    model_armor_args.AddMaliciousUriFilterSettingsEnforcement(parser)
    model_armor_args.AddPIJBFilterSettingsGroup(parser)
    model_armor_args.AddSDPFilterBasicConfigGroup(parser)
    model_armor_args.AddRaiFilterSettingsGroup(parser)
    model_armor_args.AddIntegratedServices(parser)
    model_armor_args.AddVertexAiFloorSetting(parser)
    model_armor_args.AddGoogleMcpServerFloorSetting(parser)
    model_armor_args.AddMultiLanguageDetection(parser)
    model_armor_args.AddGoogleMcpServerApiFlags(parser)

  def _GetEnumValue(self, value):
    return value.upper().replace('-', '_')

  def _GetCanonicalServiceName(self, service_name: str) -> str:
    """Returns the canonical service name, handling aliases like VERTEX_AI."""
    if service_name.upper() == 'VERTEX_AI':
      return 'AI_PLATFORM'
    return self._GetEnumValue(service_name)

  def _updateMaskForEnforcementType(
      self, *, service_floor_setting_arg, arg_enforcement_type, update_mask
  ):
    """Update the update mask for enforcement type."""
    if arg_enforcement_type == 'INSPECT_ONLY':
      update_mask.append(service_floor_setting_arg + '.inspect_only')
    elif arg_enforcement_type == 'INSPECT_AND_BLOCK':
      update_mask.append(service_floor_setting_arg + '.inspect_and_block')

  def _handleEnforcementType(self, arg_enforcement_type, service_floor_setting):
    """Handle the enforcement type for Integreated Services."""
    # Reset both oneof fields to None before setting the desired one.
    setattr(service_floor_setting, 'inspectOnly', None)
    setattr(service_floor_setting, 'inspectAndBlock', None)

    if arg_enforcement_type == 'INSPECT_ONLY':
      setattr(
          service_floor_setting,
          'inspectOnly',
          True,
      )
    elif arg_enforcement_type == 'INSPECT_AND_BLOCK':
      setattr(
          service_floor_setting,
          'inspectAndBlock',
          True,
      )

  def _updateGoogleMcpServerApis(self, service_setting, args):
    """Updates Google MCP server APIs list."""
    apis = []
    if service_setting.apis is not None:
      apis = list(service_setting.apis)

    if args.IsSpecified('google_mcp_server_apis'):
      apis = args.google_mcp_server_apis
    elif args.IsSpecified('add_google_mcp_server_apis'):
      for api in args.add_google_mcp_server_apis:
        if api not in apis:
          apis.append(api)
    elif args.IsSpecified('remove_google_mcp_server_apis'):
      for api in args.remove_google_mcp_server_apis:
        if api in apis:
          apis.remove(api)
    elif args.IsSpecified('clear_google_mcp_server_apis'):
      apis = []

    service_setting.apis = apis

  def _validateRaiFilterSettings(self, messages, filters, argument_name):
    for item in [json.loads(f) for f in filters]:
      confidence_level = item['confidenceLevel']
      filter_type = item['filterType']
      model_armor_util.ValidateEnum(
          filter_type,
          messages.RaiFilter.FilterTypeValueValuesEnum,
          argument_name,
          f'Invalid choice : {filter_type}',
      )
      model_armor_util.ValidateEnum(
          confidence_level,
          messages.RaiFilter.ConfidenceLevelValueValuesEnum,
          argument_name,
          f'Invalid choice : {confidence_level}',
      )

  def _validateIntegratedServices(self, messages, services):
    for service in services:
      canonical_service = self._GetCanonicalServiceName(service)
      model_armor_util.ValidateEnum(
          canonical_service,
          messages.FloorSetting.IntegratedServicesValueListEntryValuesEnum,
          'integrated-services',
          canonical_service,
      )

  def _validateRemoveIntegratedServices(self, messages, args):
    """Validate the arguments for removing integrated services."""
    self._validateIntegratedServices(messages, args.remove_integrated_services)
    services_config = _GetServicesConfig(messages)
    services_config_by_name = {
        service_config.name: service_config
        for service_config in services_config
    }
    for service_name in args.remove_integrated_services:
      canonical_service_name = self._GetCanonicalServiceName(service_name)
      if canonical_service_name in services_config_by_name:
        service_config = services_config_by_name[canonical_service_name]
        if canonical_service_name == service_config.name:
          if args.IsSpecified(service_config.enforcement_type_arg):
            raise exceptions.ConflictingArgumentsException(
                f'--remove-integrated-services={service_name}',
                f'--{service_config.cli_name}-enforcement-type',
            )
          if args.IsSpecified(service_config.cloud_logging_arg):
            raise exceptions.ConflictingArgumentsException(
                f'--remove-integrated-services={service_name}',
                f'--enable-{service_config.cli_name}-cloud-logging',
            )

  def _validateClearIntegratedServices(self, messages, args):
    """Validate the arguments for clearing integrated services."""
    services_config = _GetServicesConfig(messages)
    for service_config in services_config:
      if args.IsSpecified(service_config.enforcement_type_arg):
        raise exceptions.ConflictingArgumentsException(
            '--clear-integrated-services',
            f'--{service_config.cli_name}-enforcement-type',
        )
      if args.IsSpecified(service_config.cloud_logging_arg):
        raise exceptions.ConflictingArgumentsException(
            '--clear-integrated-services',
            f'--enable-{service_config.cli_name}-cloud-logging',
        )

  def _validateEnforcementType(self, enforcement_type, service_name):
    if (
        enforcement_type == 'INSPECT_ONLY'
        or enforcement_type == 'INSPECT_AND_BLOCK'
    ):
      return

    raise exceptions.InvalidArgumentException(
        f'{service_name}-enforcement-type',
        f'argument --{service_name}-enforcement-type: Invalid choice:'
        f' {enforcement_type!r}. Valid choices are [INSPECT_AND_BLOCK,'
        ' INSPECT_ONLY].',
    )

  def _validateArgs(self, messages, args):
    """Validate the arguments."""
    if args.pi_and_jailbreak_filter_settings_confidence_level:
      model_armor_util.ValidateEnum(
          args.pi_and_jailbreak_filter_settings_confidence_level,
          messages.PiAndJailbreakFilterResult.ConfidenceLevelValueValuesEnum,
          'pi-and-jailbreak-filter-settings-confidence-level',
          'Invalid choice :'
          f' {args.pi_and_jailbreak_filter_settings_confidence_level}',
      )
    if args.pi_and_jailbreak_filter_settings_enforcement:
      model_armor_util.ValidateEnum(
          args.pi_and_jailbreak_filter_settings_enforcement,
          messages.PiAndJailbreakFilterSettings.FilterEnforcementValueValuesEnum,
          'pi-and-jailbreak-filter-settings-enforcement',
          'Invalid choice :'
          f' {args.pi_and_jailbreak_filter_settings_enforcement}',
      )
    if args.malicious_uri_filter_settings_enforcement:
      model_armor_util.ValidateEnum(
          args.malicious_uri_filter_settings_enforcement,
          messages.MaliciousUriFilterSettings.FilterEnforcementValueValuesEnum,
          'malicious-uri-filter-settings-enforcement',
          f'Invalid choice : {args.malicious_uri_filter_settings_enforcement}',
      )
    if args.basic_config_filter_enforcement:
      model_armor_util.ValidateEnum(
          args.basic_config_filter_enforcement,
          messages.SdpBasicConfig.FilterEnforcementValueValuesEnum,
          'basic-config-filter-enforcement',
          f'Invalid choice : {args.basic_config_filter_enforcement}',
      )
    if args.rai_settings_filters:
      self._validateRaiFilterSettings(
          messages, args.rai_settings_filters, 'rai-settings-filters'
      )
    if args.add_rai_settings_filters:
      self._validateRaiFilterSettings(
          messages, args.add_rai_settings_filters, 'add-rai-settings-filters'
      )
    if args.remove_rai_settings_filters:
      self._validateRaiFilterSettings(
          messages,
          args.remove_rai_settings_filters,
          'remove-rai-settings-filters',
      )
    if args.IsSpecified('basic_config_filter_enforcement') and ((
        args.IsSpecified('advanced_config_inspect_template')
        or args.IsSpecified('advanced_config_deidentify_template')
    )):
      raise exceptions.ConflictingArgumentsException(
          'basic_config_filter_enforcement', 'sdp_advanced_config_*_template'
      )
    if args.IsSpecified('add_integrated_services'):
      self._validateIntegratedServices(messages, args.add_integrated_services)
    if args.IsSpecified('remove_integrated_services'):
      self._validateRemoveIntegratedServices(messages, args)
    if args.IsSpecified('clear_integrated_services'):
      self._validateClearIntegratedServices(messages, args)
    if args.IsSpecified('vertex_ai_enforcement_type'):
      self._validateEnforcementType(
          args.vertex_ai_enforcement_type, 'vertex-ai'
      )
    if args.IsSpecified('google_mcp_server_enforcement_type'):
      self._validateEnforcementType(
          args.google_mcp_server_enforcement_type, 'google-mcp-server'
      )

  def _RunUpdate(self, original, args):
    api_version = model_armor_api.GetApiFromTrack(self.ReleaseTrack())
    api_client = model_armor_api.FloorSettings(api_version=api_version)
    messages = api_client.GetMessages()
    self._validateArgs(messages, args)
    floor_setting_updated = original
    defaulted_enforcement = []

    if floor_setting_updated.filterConfig is None:
      floor_setting_updated.filterConfig = messages.FilterConfig()

    # Collect the list of update masks
    update_mask = []
    if args.IsSpecified(
        'pi_and_jailbreak_filter_settings_enforcement'
    ) or args.IsSpecified('pi_and_jailbreak_filter_settings_confidence_level'):
      update_mask.append('filter_config.pi_and_jailbreak_filter_settings')
    if args.IsSpecified('malicious_uri_filter_settings_enforcement'):
      update_mask.append('filter_config.malicious_uri_filter_settings')
    if (
        args.IsSpecified('basic_config_filter_enforcement')
        or args.IsSpecified('advanced_config_inspect_template')
        or args.IsSpecified('advanced_config_deidentify_template')
    ):
      update_mask.append('filter_config.sdp_settings')
    if (
        args.IsSpecified('add_rai_settings_filters')
        or args.IsSpecified('remove_rai_settings_filters')
        or args.IsSpecified('clear_rai_settings_filters')
        or args.IsSpecified('rai_settings_filters')
    ):
      update_mask.append('filter_config.rai_settings')
    if args.IsSpecified('enable_floor_setting_enforcement'):
      update_mask.append('enable_floor_setting_enforcement')
    if (
        args.IsSpecified('add_integrated_services')
        or args.IsSpecified('remove_integrated_services')
        or args.IsSpecified('clear_integrated_services')
    ):
      update_mask.append('integrated_services')
    if args.IsSpecified('vertex_ai_enforcement_type'):
      self._updateMaskForEnforcementType(
          service_floor_setting_arg='ai_platform_floor_setting',
          arg_enforcement_type=args.vertex_ai_enforcement_type,
          update_mask=update_mask,
      )
    if args.IsSpecified('google_mcp_server_enforcement_type'):
      self._updateMaskForEnforcementType(
          service_floor_setting_arg='google_mcp_server_floor_setting',
          arg_enforcement_type=args.google_mcp_server_enforcement_type,
          update_mask=update_mask,
      )
    if args.IsSpecified('enable_vertex_ai_cloud_logging'):
      update_mask.append('ai_platform_floor_setting.enable_cloud_logging')
    if args.IsSpecified('enable_google_mcp_server_cloud_logging'):
      update_mask.append('google_mcp_server_floor_setting.enable_cloud_logging')
    if (
        args.IsSpecified('google_mcp_server_apis')
        or args.IsSpecified('add_google_mcp_server_apis')
        or args.IsSpecified('remove_google_mcp_server_apis')
        or args.IsSpecified('clear_google_mcp_server_apis')
    ):
      update_mask.append('google_mcp_server_floor_setting.apis')
    if args.IsSpecified('enable_multi_language_detection'):
      update_mask.append('floor_setting_metadata.multi_language_detection')

    if not update_mask:
      possible_args = [
          '--pi-and-jailbreak-filter-settings-enforcement',
          '--pi-and-jailbreak-filter-settings-confidence-level',
          '--malicious-uri-filter-settings-enforcement',
          '--basic-config-filter-enforcement',
          '--advanced-config-inspect-template',
          '--advanced-config-deidentify-template',
          '--rai-settings-filters',
          '--add-rai-settings-filters',
          '--remove-rai-settings-filters',
          '--clear-rai-settings-filters',
          '--enable-floor-setting-enforcement',
          '--add-integrated-services',
          '--remove-integrated-services',
          '--clear-integrated-services',
          '--enable-vertex-ai-cloud-logging',
          '--vertex-ai-enforcement-type',
          '--enable-google-mcp-server-cloud-logging',
          '--google-mcp-server-enforcement-type',
          '--enable-multi-language-detection',
          '--google-mcp-server-apis',
          '--add-google-mcp-server-apis',
          '--remove-google-mcp-server-apis',
          '--clear-google-mcp-server-apis',
      ]

      raise exceptions.MinimumArgumentException(
          possible_args,
          self.NO_CHANGES_MESSAGE.format(floor_setting=args.full_uri),
      )

    if 'filter_config.rai_settings' in update_mask:
      self._UpdateRaiSettings(messages, args, floor_setting_updated)

    if 'filter_config.pi_and_jailbreak_filter_settings' in update_mask:
      self._UpdatePiAndJailbreakSettings(messages, args, floor_setting_updated)

    if 'filter_config.malicious_uri_filter_settings' in update_mask:
      self._UpdateMaliciousUriSettings(messages, args, floor_setting_updated)

    if 'filter_config.sdp_settings' in update_mask:
      self._UpdateSdpSettings(messages, args, floor_setting_updated)

    if (
        args.IsSpecified('add_integrated_services')
        or args.IsSpecified('remove_integrated_services')
        or args.IsSpecified('clear_integrated_services')
    ):
      default_enforcement = self._UpdateIntegratedServices(
          messages, args, floor_setting_updated, update_mask
      )
      defaulted_enforcement.extend(default_enforcement.keys())

    # Handle Service Specific FloorSettings
    # (e.g., Cloud Logging, Enforcement Type).
    self._UpdateServiceSpecificSettings(messages, args, floor_setting_updated)

    if args.IsSpecified('enable_multi_language_detection'):
      self._UpdateMultiLanguageDetection(messages, args, floor_setting_updated)

    if 'enable_floor_setting_enforcement' in update_mask:
      if args.IsSpecified('enable_floor_setting_enforcement'):
        floor_setting_updated.enableFloorSettingEnforcement = (
            args.enable_floor_setting_enforcement.lower() == 'true'
        )
    return (
        api_client.Update(
            name=args.full_uri,
            floor_setting=floor_setting_updated,
            update_mask=update_mask,
        ),
        defaulted_enforcement,
    )

  def _UpdateRaiSettings(self, messages, args, floor_setting_updated):
    """Updates the RaiFilterSettings in the FloorSetting."""
    rai_filters = []
    if floor_setting_updated.filterConfig.raiSettings is not None:
      rai_filters = floor_setting_updated.filterConfig.raiSettings.raiFilters

    if args.IsSpecified('rai_settings_filters'):
      rai_filters = []
      for item in [json.loads(f) for f in args.rai_settings_filters]:
        arg_filter_type = messages.RaiFilter.FilterTypeValueValuesEnum(
            self._GetEnumValue(item['filterType'])
        )
        arg_confidence_level = (
            messages.RaiFilter.ConfidenceLevelValueValuesEnum(
                self._GetEnumValue(item['confidenceLevel'])
            )
        )
        rai_filters.append(
            messages.RaiFilter(
                filterType=arg_filter_type,
                confidenceLevel=arg_confidence_level,
            )
        )
      floor_setting_updated.filterConfig.raiSettings = (
          messages.RaiFilterSettings(raiFilters=rai_filters)
      )
    if args.IsSpecified('add_rai_settings_filters'):
      for item in [json.loads(f) for f in args.add_rai_settings_filters]:
        already_exists = False
        arg_filter_type = messages.RaiFilter.FilterTypeValueValuesEnum(
            self._GetEnumValue(item['filterType'])
        )
        arg_confidence_level = (
            messages.RaiFilter.ConfidenceLevelValueValuesEnum(
                self._GetEnumValue(item['confidenceLevel'])
            )
        )
        for _, rai_filter in enumerate(rai_filters):
          if arg_filter_type == rai_filter.filterType:
            already_exists = True
            rai_filters.remove(rai_filter)
            rai_filters.append(
                messages.RaiFilter(
                    filterType=arg_filter_type,
                    confidenceLevel=arg_confidence_level,
                )
            )
            break
        if not already_exists:
          rai_filters.append(
              messages.RaiFilter(
                  filterType=arg_filter_type,
                  confidenceLevel=arg_confidence_level,
              )
          )
        floor_setting_updated.filterConfig.raiSettings = (
            messages.RaiFilterSettings(raiFilters=rai_filters)
        )

    if args.IsSpecified('remove_rai_settings_filters'):
      for item in [json.loads(f) for f in args.remove_rai_settings_filters]:
        arg_filter_type = messages.RaiFilter.FilterTypeValueValuesEnum(
            self._GetEnumValue(item['filterType'])
        )
        arg_confidence_level = (
            messages.RaiFilter.ConfidenceLevelValueValuesEnum(
                self._GetEnumValue(item['confidenceLevel'])
            )
        )
        for _, rai_filter in enumerate(rai_filters):
          if (
              messages.RaiFilter.FilterTypeValueValuesEnum(arg_filter_type)
              == rai_filter.filterType
              and messages.RaiFilter.ConfidenceLevelValueValuesEnum(
                  arg_confidence_level
              )
              == rai_filter.confidenceLevel
          ):
            rai_filters.remove(rai_filter)
            break
    if args.IsSpecified('clear_rai_settings_filters'):
      floor_setting_updated.filterConfig.raiSettings = None

  def _UpdatePiAndJailbreakSettings(
      self, messages, args, floor_setting_updated
  ):
    """Updates the PiAndJailbreakFilterSettings in the FloorSetting."""
    if args.IsSpecified('pi_and_jailbreak_filter_settings_enforcement'):
      pi_and_jailbreak_filter_settings_enforcement = self._GetEnumValue(
          args.pi_and_jailbreak_filter_settings_enforcement
      )
      if (
          floor_setting_updated.filterConfig.piAndJailbreakFilterSettings
          is None
      ):
        floor_setting_updated.filterConfig.piAndJailbreakFilterSettings = messages.PiAndJailbreakFilterSettings(
            filterEnforcement=messages.PiAndJailbreakFilterSettings.FilterEnforcementValueValuesEnum(
                pi_and_jailbreak_filter_settings_enforcement
            )
        )
      else:
        floor_setting_updated.filterConfig.piAndJailbreakFilterSettings.filterEnforcement = messages.PiAndJailbreakFilterSettings.FilterEnforcementValueValuesEnum(
            pi_and_jailbreak_filter_settings_enforcement
        )
    if args.IsSpecified('pi_and_jailbreak_filter_settings_confidence_level'):
      pi_and_jailbreak_filter_settings_confidence_level = self._GetEnumValue(
          args.pi_and_jailbreak_filter_settings_confidence_level
      )
      if (
          floor_setting_updated.filterConfig.piAndJailbreakFilterSettings
          is None
      ):
        floor_setting_updated.filterConfig.piAndJailbreakFilterSettings = messages.PiAndJailbreakFilterSettings(
            confidenceLevel=messages.PiAndJailbreakFilterSettings.ConfidenceLevelValueValuesEnum(
                pi_and_jailbreak_filter_settings_confidence_level
            )
        )
      else:
        floor_setting_updated.filterConfig.piAndJailbreakFilterSettings.confidenceLevel = messages.PiAndJailbreakFilterSettings.ConfidenceLevelValueValuesEnum(
            pi_and_jailbreak_filter_settings_confidence_level
        )

  def _UpdateMultiLanguageDetection(
      self, messages, args, floor_setting_updated
  ):
    """Updates the MultiLanguageDetection in the FloorSetting."""
    if floor_setting_updated.floorSettingMetadata is None:
      floor_setting_updated.floorSettingMetadata = (
          messages.FloorSettingMetadata()
      )
    if (
        floor_setting_updated.floorSettingMetadata.multiLanguageDetection
        is None
    ):
      floor_setting_updated.floorSettingMetadata.multiLanguageDetection = (
          messages.FloorSettingFloorSettingMetadataMultiLanguageDetection(
              enableMultiLanguageDetection=args.enable_multi_language_detection
          )
      )
    else:
      (
          floor_setting_updated.floorSettingMetadata.multiLanguageDetection.enableMultiLanguageDetection
      ) = args.enable_multi_language_detection

  def _UpdateServiceSpecificSettings(
      self, messages, args, floor_setting_updated
  ):
    """Handles Cloud Logging and Enforcement Type for Integrated Services.

    Also handles API list for GOOGLE_MCP_SERVER.

    Args:
      messages: The messages module for the API version.
      args: The parsed arguments for the command.
      floor_setting_updated: The floor setting resource to update.
    """
    services_config = _GetServicesConfig(messages)
    for service_config in services_config:
      service_attr = service_config.service_floor_setting_attr
      service_setting = getattr(floor_setting_updated, service_attr)
      is_cloud_logging_specified = args.IsSpecified(
          service_config.cloud_logging_arg
      )
      is_enforcement_type_specified = args.IsSpecified(
          service_config.enforcement_type_arg
      )
      is_google_mcp_server_apis_specified = False
      if service_config.name == 'GOOGLE_MCP_SERVER':
        is_google_mcp_server_apis_specified = (
            args.IsSpecified('google_mcp_server_apis')
            or args.IsSpecified('add_google_mcp_server_apis')
            or args.IsSpecified('remove_google_mcp_server_apis')
            or args.IsSpecified('clear_google_mcp_server_apis')
        )
      if (
          is_cloud_logging_specified
          or is_enforcement_type_specified
          or is_google_mcp_server_apis_specified
      ):
        if service_setting is None:
          service_setting = service_config.setting_message()
          setattr(floor_setting_updated, service_attr, service_setting)

      if is_cloud_logging_specified:
        service_setting = getattr(floor_setting_updated, service_attr)
        service_setting.enableCloudLogging = getattr(
            args, service_config.cloud_logging_arg
        )

      if is_enforcement_type_specified:
        arg_enforcement_type = self._GetEnumValue(
            getattr(args, service_config.enforcement_type_arg)
        )
        self._handleEnforcementType(arg_enforcement_type, service_setting)

      if is_google_mcp_server_apis_specified:
        self._updateGoogleMcpServerApis(service_setting, args)

  def _UpdateIntegratedServices(
      self, messages, args, floor_setting_updated, update_mask
  ):
    """Updates the IntegratedServices in the FloorSetting."""
    defaulted_enforcement = {}
    services_config = _GetServicesConfig(messages)
    if args.IsSpecified('clear_integrated_services'):
      for service_config in services_config:
        service_attr = service_config.service_floor_setting_attr
        if (
            getattr(floor_setting_updated, service_attr) is not None
            and service_config.enum in floor_setting_updated.integratedServices
        ):
          setattr(floor_setting_updated, service_attr, None)
          update_mask.append(service_config.service_floor_setting_arg)
      floor_setting_updated.integratedServices = []
    if args.IsSpecified('add_integrated_services') or args.IsSpecified(
        'remove_integrated_services'
    ):
      integrated_services = []
      if floor_setting_updated.integratedServices is not None:
        integrated_services = list(floor_setting_updated.integratedServices)

      services_config_by_name = {
          service_config.name: service_config
          for service_config in services_config
      }
      if args.IsSpecified('add_integrated_services'):
        for service in args.add_integrated_services:
          arg_service = self._GetCanonicalServiceName(service)
          if arg_service in services_config_by_name:
            found_config = services_config_by_name[arg_service]
            if found_config.enum in integrated_services:
              continue
            # if no enforcement type is specified, set it to INSPECT_ONLY.
            if not args.IsSpecified(found_config.enforcement_type_arg):
              setting_attr = found_config.service_floor_setting_attr
              service_setting = getattr(
                  floor_setting_updated,
                  setting_attr,
              )
              if service_setting is None:
                service_setting = found_config.setting_message()
                setattr(
                    floor_setting_updated,
                    setting_attr,
                    service_setting,
                )
              self._handleEnforcementType('INSPECT_ONLY', service_setting)
              self._updateMaskForEnforcementType(
                  service_floor_setting_arg=found_config.service_floor_setting_arg,
                  arg_enforcement_type='INSPECT_ONLY',
                  update_mask=update_mask,
              )
              setattr(
                  floor_setting_updated,
                  setting_attr,
                  service_setting,
              )
              defaulted_enforcement[found_config.enforcement_type_flag_name] = (
                  True
              )
            if found_config.enum not in integrated_services:
              integrated_services.append(found_config.enum)

      if args.IsSpecified('remove_integrated_services'):
        for service in args.remove_integrated_services:
          arg_service = self._GetCanonicalServiceName(service)
          if arg_service in services_config_by_name:
            found_config = services_config_by_name[arg_service]
            if found_config.enum in integrated_services:
              integrated_services.remove(found_config.enum)
              setattr(
                  floor_setting_updated,
                  found_config.service_floor_setting_attr,
                  None,
              )
              update_mask.append(found_config.service_floor_setting_arg)

      floor_setting_updated.integratedServices = integrated_services
    return defaulted_enforcement

  def _UpdateMaliciousUriSettings(self, messages, args, floor_setting_updated):
    """Updates the MaliciousUriFilterSettings in the FloorSetting."""
    if args.IsSpecified('malicious_uri_filter_settings_enforcement'):
      if floor_setting_updated.filterConfig.maliciousUriFilterSettings is None:
        floor_setting_updated.filterConfig.maliciousUriFilterSettings = messages.MaliciousUriFilterSettings(
            filterEnforcement=messages.MaliciousUriFilterSettings.FilterEnforcementValueValuesEnum(
                self._GetEnumValue(
                    args.malicious_uri_filter_settings_enforcement
                )
            )
        )
      else:
        floor_setting_updated.filterConfig.maliciousUriFilterSettings.filterEnforcement = messages.MaliciousUriFilterSettings.FilterEnforcementValueValuesEnum(
            self._GetEnumValue(args.malicious_uri_filter_settings_enforcement)
        )

  def _UpdateSdpSettings(self, messages, args, floor_setting_updated):
    """Updates the SdpFilterSettings in the FloorSetting."""
    if args.IsSpecified('basic_config_filter_enforcement'):
      if floor_setting_updated.filterConfig.sdpSettings is None:
        floor_setting_updated.filterConfig.sdpSettings = messages.SdpFilterSettings(
            basicConfig=messages.SdpBasicConfig(
                filterEnforcement=(
                    messages.SdpBasicConfig.FilterEnforcementValueValuesEnum(
                        self._GetEnumValue(args.basic_config_filter_enforcement)
                    )
                )
            )
        )
      else:
        floor_setting_updated.filterConfig.sdpSettings.advancedConfig = None
        floor_setting_updated.filterConfig.sdpSettings.basicConfig = messages.SdpBasicConfig(
            filterEnforcement=messages.SdpBasicConfig.FilterEnforcementValueValuesEnum(
                self._GetEnumValue(args.basic_config_filter_enforcement)
            )
        )
    if args.IsSpecified('advanced_config_inspect_template'):
      if floor_setting_updated.filterConfig.sdpSettings is None:
        floor_setting_updated.filterConfig.sdpSettings = (
            messages.SdpFilterSettings(
                advancedConfig=messages.SdpAdvancedConfig(
                    inspectTemplate=args.advanced_config_inspect_template
                )
            )
        )
      else:
        floor_setting_updated.filterConfig.sdpSettings.basicConfig = None
        if (
            floor_setting_updated.filterConfig.sdpSettings.advancedConfig
            is None
        ):
          floor_setting_updated.filterConfig.sdpSettings.advancedConfig = (
              messages.SdpAdvancedConfig(
                  inspectTemplate=args.advanced_config_inspect_template
              )
          )
        else:
          floor_setting_updated.filterConfig.sdpSettings.advancedConfig.inspectTemplate = (
              args.advanced_config_inspect_template
          )
    if args.IsSpecified('advanced_config_deidentify_template'):
      if floor_setting_updated.filterConfig.sdpSettings is None:
        floor_setting_updated.filterConfig.sdpSettings = (
            messages.SdpFilterSettings(
                advancedConfig=messages.SdpAdvancedConfig(
                    deidentifyTemplate=args.advanced_config_deidentify_template
                )
            )
        )
      else:
        floor_setting_updated.filterConfig.sdpSettings.basicConfig = None
        if (
            floor_setting_updated.filterConfig.sdpSettings.advancedConfig
            is None
        ):
          floor_setting_updated.filterConfig.sdpSettings.advancedConfig = (
              messages.SdpAdvancedConfig(
                  deidentifyTemplate=args.advanced_config_deidentify_template
              )
          )
        else:
          floor_setting_updated.filterConfig.sdpSettings.advancedConfig.deidentifyTemplate = (
              args.advanced_config_deidentify_template
          )

  def Run(self, args):
    api_version = model_armor_api.GetApiFromTrack(self.ReleaseTrack())
    full_uri = args.full_uri
    original = model_armor_api.FloorSettings(api_version=api_version).Get(
        full_uri
    )
    res, defaulted_enforcement = self._RunUpdate(original, args)
    defaulted_enforcement_services_flags = set()
    if defaulted_enforcement:
      for service in defaulted_enforcement:
        defaulted_enforcement_services_flags.add(f'{service}=INSPECT_AND_BLOCK')

      sorted_flags = sorted(defaulted_enforcement_services_flags)
      defaulted_services_str = ' and '.join(
          f'"{service}"' for service in sorted_flags
      )
      message = (
          'Enforcement type defaulted to INSPECT_ONLY. This means that traffic'
          ' violating Model Armor settings will be inspected but not blocked.'
          f' To block such traffic, please specify {defaulted_services_str}.'
      )
      log.status.Print(message)
    return res



# -*- coding: utf-8 -*- #
# Copyright 2015 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Supplementary help for gcloud startup options."""


from googlecloudsdk.calliope import base


@base.UniverseCompatible
class Startup(base.TopicCommand):
  """Supplementary help for gcloud startup options.


  This page provides supplementary help for configuring the gcloud command-line
  tool's startup behavior, including Python interpreter selection and
  environment variables.

  ## CONFIGURING THE PYTHON INTERPRETER

  The `gcloud` CLI requires a compatible Python version (3.10-3.14) to run. In
  most gcloud CLI installations, the gcloud installer manages the Python
  installation (version 3.14) for the user. Configuring the Python
  interpreter is only supported in specific scenarios described below.

  ### When to Use `CLOUDSDK_PYTHON` environment variable

  You can ONLY consider using the CLOUDSDK_PYTHON environment variable when you
  are on:

  - *Linux*: Installing from a versioned archives (`*.tar.gz*`) on ARM or x86
  architectures where Python is NOT included as part of the installation, gcloud
  will search for a compatible Python on your system PATH (looking for python3
  then python). You can set the `CLOUDSDK_PYTHON` environment variable to the
  full path of your preferred compatible Python interpreter if the default one
  found on PATH is not desired.
  - *macOS*: Available only for archived install and where the Python version
  matches the Python version managed by the gcloud CLI Tools install. Note
  that this is not available for Homebrew installations of gcloud CLI Tools.
  - *Windows*: gcloud installer includes the required Python interpreter by
  default. Only use `CLOUDSDK_PYTHON` if you need to use a different
  Python installation.

  Example Usages:

    # Use the python3 interpreter on your path
    $ export CLOUDSDK_PYTHON=python3

    # Use a python you have installed in a special location
    $ export CLOUDSDK_PYTHON=/usr/local/my-custom-python-install/python

  ### Other Components

  `gsutil` versions 5.0 and later support Python 3.10-3.14. To use a different
  interpreter for `gsutil` than for the other Python tools, set the
  `CLOUDSDK_GSUTIL_PYTHON` environment variable to the interpreter that you
  want.

  `bq` versions 2.0.99 and later support Python 3.10-3.14. To use a different
  interpreter for `bq` than for the other Python tools, set the
  `CLOUDSDK_BQ_PYTHON` environment variable to the interpreter that you want.

  ## SETTING CONFIGURATIONS AND PROPERTIES

  Your active configuration can also be set via the environment variable
  `CLOUDSDK_ACTIVE_CONFIG_NAME`. This allows you to specify a certain
  configuration in a given terminal session without changing the global
  default configuration.

  In addition to being able to set them via `gcloud config set`,
  each `gcloud` property has a corresponding environment variable. They take
  the form: `CLOUDSDK_SECTION_PROPERTY`. For example, if you wanted to
  change your active project for just one terminal you could run:

    $ export CLOUDSDK_CORE_PROJECT=my-project

  For more information, see `gcloud topic configurations`.
  """



# -*- coding: utf-8 -*- #
# Copyright 2026 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Command to delete an image streaming cache."""

from googlecloudsdk.api_lib.artifacts import exceptions as ar_exceptions
from googlecloudsdk.calliope import base
from googlecloudsdk.command_lib.artifacts import flags
from googlecloudsdk.command_lib.artifacts import requests as ar_requests
from googlecloudsdk.command_lib.util.concepts import concept_parsers
from googlecloudsdk.command_lib.util.concepts import presentation_specs
from googlecloudsdk.core import log


@base.ReleaseTracks(base.ReleaseTrack.GA)
@base.DefaultUniverseOnly
class Delete(base.DeleteCommand):
  """Removes the prewarmed state of a specific artifact from a specific stream location."""

  detailed_help = {
      "DESCRIPTION": """
          {description}
      """,
      "EXAMPLES": """
          # Removes the prewarmed state of a specific artifact using flags
          $ {command} --package=my-pkg --tag=latest --repository=my-repo --location=us-central1 --stream-location=us-east1

          # Using full resource name
          $ {command} projects/my-proj/locations/us-central1/repositories/my-repo/packages/my-pkg/versions/sha256:abc123 --stream-location=us-east1
      """,
  }

  @staticmethod
  def Args(parser):
    flags.GetStreamLocationArg().AddToParser(parser)

    resource_group = parser.add_group(mutex=True, required=True)
    resource_group.add_argument(
        "ARTIFACT",
        nargs="?",
        help="The full resource name of the version or tag to remove.",
    )

    flag_group = resource_group.add_group()

    concept_parsers.ConceptParser(
        [
            presentation_specs.ResourcePresentationSpec(
                "--package",
                flags.GetPackageResourceSpec(),
                "The package to remove.",
                group=flag_group,
            ),
        ],
        command_level_fallthroughs={},
    ).AddToParser(parser)

    # Mutex group for EITHER --version or --tag ID
    version_or_tag_group = flag_group.add_group(mutex=True, required=True)
    version_or_tag_group.add_argument(
        "--version",
        help="The version to remove (e.g., 'sha256:abc123').",
    )
    version_or_tag_group.add_argument(
        "--tag",
        help="The tag to remove (e.g., 'latest').",
    )

  def Run(self, args):
    """This is what gets called when the user runs this command.

    Args:
      args: an argparse namespace. All the arguments that were provided to this
        command invocation.

    Returns:
      The image streaming cache deleted.
    """
    client = ar_requests.GetClient()
    messages = ar_requests.GetMessages()

    if args.ARTIFACT:
      artifact_ref = args.ARTIFACT
    else:
      package_concept = args.CONCEPTS.package.Parse()
      if not package_concept:
        raise ar_exceptions.InvalidInputValueError(
            "Missing required flags for the package (e.g., --package,"
            " --repository, --location)."
        )

      package_name = package_concept.RelativeName()

      if args.IsSpecified("version"):
        artifact_ref = f"{package_name}/versions/{args.version}"
      elif args.IsSpecified("tag"):
        artifact_ref = f"{package_name}/tags/{args.tag}"
      else:
        raise ar_exceptions.InvalidInputValueError(
            "Either --version or --tag must be specified with component flags."
        )

    result = ar_requests.RemovePrewarmedArtifact(
        client,
        messages,
        artifact_ref,
        args.stream_location,
    )
    log.status.Print(
        "Successfully removed image streaming cache for [{}] in [{}].".format(
            artifact_ref, args.stream_location
        )
    )
    return result



# -*- coding: utf-8 -*- #
# Copyright 2026 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Command for starting instances."""

import argparse

from googlecloudsdk.calliope import base
from googlecloudsdk.command_lib.run import connection_context
from googlecloudsdk.command_lib.run import flags
from googlecloudsdk.command_lib.run import pretty_print
from googlecloudsdk.command_lib.run import resource_args
from googlecloudsdk.command_lib.run import serverless_operations
from googlecloudsdk.command_lib.util.concepts import concept_parsers
from googlecloudsdk.command_lib.util.concepts import presentation_specs


@base.UniverseCompatible
@base.RegionalEndpointsSupported
class Start(base.Command):
  """Start a stopped instance."""

  detailed_help = {
      'DESCRIPTION': (
          """
          {description}
          """
      ),
      'EXAMPLES': (
          """
          To start an instance:

              $ {command} my-instance
          """
      ),
  }

  @staticmethod
  def CommonArgs(parser: argparse.ArgumentParser) -> None:
    instance_presentation = presentation_specs.ResourcePresentationSpec(
        'INSTANCE',
        resource_args.GetInstanceResourceSpec(),
        'Instance to start.',
        required=True,
        prefixes=False,
    )
    flags.AddAsyncFlag(parser)
    concept_parsers.ConceptParser([instance_presentation]).AddToParser(parser)

  @staticmethod
  def Args(parser: argparse.ArgumentParser) -> None:
    Start.CommonArgs(parser)

  def Run(self, args: argparse.Namespace) -> None:
    """Start an instance."""
    instance_ref = args.CONCEPTS.instance.Parse()
    conn_context = connection_context.GetConnectionContext(
        args, flags.Product.RUN, self.ReleaseTrack()
    )
    with serverless_operations.Connect(conn_context) as client:
      instance = client.StartInstance(
          instance_ref, asyn=args.async_
      )

    if args.async_:
      pretty_print.Success(
          f'Instance [{instance.name}] is being started.'
      )
    else:
      pretty_print.Success(
          f'Started instance [{instance.name}].'
      )



# -*- coding: utf-8 -*- #
# Copyright 2026 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Client for interacting with Cloud FTP API."""

from apitools.base.py import list_pager
from googlecloudsdk.api_lib.util import apis as core_apis


class FtpClient(object):
  """Client wrapper for Cloud FTP API.

  Attributes:
    client: The API client for interacting with the FTP service.
    messages: The API messages module for the FTP service.
    servers_service: The service for interacting with Server resources.
    users_service: The service for interacting with User resources.
    operations_service: Service for long-running operations.
  """

  def __init__(self, api_version='v1alpha'):
    """Initializes the instance."""
    self.client = core_apis.GetClientInstance('ftp', api_version)
    self.messages = core_apis.GetMessagesModule('ftp', api_version)
    self.servers_service = self.client.projects_locations_servers
    self.users_service = self.client.projects_locations_servers_users
    self.operations_service = self.client.projects_locations_operations

  # --- Server Methods ---

  def CreateServer(self, parent, server_id, server_msg):
    """Calls create server API."""
    req = self.messages.FtpProjectsLocationsServersCreateRequest(
        parent=parent,
        serverId=server_id,
        server=server_msg,
    )
    return self.servers_service.Create(req)

  def GetServer(self, name):
    """Calls get server API."""
    req = self.messages.FtpProjectsLocationsServersGetRequest(name=name)
    return self.servers_service.Get(req)

  def UpdateServer(self, server_msg, update_mask):
    """Calls patch server API."""
    req = self.messages.FtpProjectsLocationsServersPatchRequest(
        name=server_msg.name,
        server=server_msg,
        updateMask=','.join(update_mask),
    )
    return self.servers_service.Patch(req)

  def DeleteServer(self, name):
    """Calls delete server API."""
    req = self.messages.FtpProjectsLocationsServersDeleteRequest(name=name)
    return self.servers_service.Delete(req)

  def ListServers(self, parent, page_size=None, limit=None):
    """Calls list servers API."""
    req = self.messages.FtpProjectsLocationsServersListRequest(parent=parent)
    return list_pager.YieldFromList(
        self.servers_service,
        req,
        batch_size=page_size,
        limit=limit,
        batch_size_attribute='pageSize',
        field='servers',
    )

  # --- User Methods ---

  def CreateUser(self, parent, user_id, user_msg):
    """Calls create user API."""
    req = self.messages.FtpProjectsLocationsServersUsersCreateRequest(
        parent=parent,
        userId=user_id,
        user=user_msg,
    )
    return self.users_service.Create(req)

  def GetUser(self, name):
    """Calls get user API."""
    req = self.messages.FtpProjectsLocationsServersUsersGetRequest(name=name)
    return self.users_service.Get(req)

  def UpdateUser(self, user_msg, update_mask):
    """Calls patch user API."""
    req = self.messages.FtpProjectsLocationsServersUsersPatchRequest(
        name=user_msg.name,
        user=user_msg,
        updateMask=','.join(update_mask),
    )
    return self.users_service.Patch(req)

  def DeleteUser(self, name, force=False):
    """Calls delete user API."""
    req = self.messages.FtpProjectsLocationsServersUsersDeleteRequest(
        name=name, force=force
    )
    return self.users_service.Delete(req)

  def ListUsers(self, parent, page_size=None, limit=None):
    """Calls list users API."""
    req = self.messages.FtpProjectsLocationsServersUsersListRequest(
        parent=parent
    )
    return list_pager.YieldFromList(
        self.users_service,
        req,
        batch_size=page_size,
        limit=limit,
        batch_size_attribute='pageSize',
        field='users',
    )

  # --- Operation Methods ---

  def GetOperation(self, name):
    """Calls get operation API."""
    req = self.messages.FtpProjectsLocationsOperationsGetRequest(name=name)
    return self.operations_service.Get(req)



# -*- coding: utf-8 -*- #
# Copyright 2024 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Describe the AutokeyConfig of a folder or project."""

from googlecloudsdk.api_lib.cloudkms import base as cloudkms_base
from googlecloudsdk.calliope import base
from googlecloudsdk.command_lib.kms import flags


@base.UniverseCompatible
class Describe(base.DescribeCommand):
  r"""Describe the AutokeyConfig of a folder or project.

  {command} can be used to retrieve the AutokeyConfig of a folder or project.

  ## EXAMPLES

  The following command retrieves the AutokeyConfig of a folder having folder-id
  `123`:

    $ {command} --folder=123

  The following command retrieves the AutokeyConfig of a project having
  project-id `my-project`:

    $ {command} --project=my-project
  """

  @staticmethod
  def Args(parser):
    flags.AddAutokeyConfigResourceFlags(parser)

  def Run(self, args):
    client = cloudkms_base.GetClientInstance()
    messages = client.MESSAGES_MODULE

    if args.folder:
      autokey_config_name = f'folders/{args.folder}/autokeyConfig'
      return client.folders.GetAutokeyConfig(
          messages.CloudkmsFoldersGetAutokeyConfigRequest(
              name=autokey_config_name
          )
      )

    # Otherwise, it is a project.
    autokey_config_name = f'projects/{args.project}/autokeyConfig'
    return client.projects.GetAutokeyConfig(
        messages.CloudkmsProjectsGetAutokeyConfigRequest(
            name=autokey_config_name
        )
    )



# -*- coding: utf-8 -*- #
# Copyright 2024 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Creates Backup and DR Backup Plan Association."""



from apitools.base.py import exceptions as apitools_exceptions
from googlecloudsdk.api_lib.backupdr import util
from googlecloudsdk.api_lib.backupdr.backup_plan_associations import BackupPlanAssociationsClient
from googlecloudsdk.api_lib.util import exceptions
from googlecloudsdk.calliope import base
from googlecloudsdk.command_lib.backupdr import flags
from googlecloudsdk.core import log


@base.DefaultUniverseOnly
@base.ReleaseTracks(
    base.ReleaseTrack.ALPHA, base.ReleaseTrack.BETA, base.ReleaseTrack.GA
)
class Create(base.CreateCommand):
  """Create a new backup plan association."""

  detailed_help = {
      'BRIEF': 'Creates a new backup plan association',
      'DESCRIPTION': (
          'Create a new backup plan association in the project. It can only be'
          ' created in locations where Backup and DR is available.'
      ),
      'EXAMPLES': """\
        To create a new backup plan association `sample-bpa` in project `sample-project` and location `us-central1` for resource `sample-resource-uri` with backup plan `sample-backup-plan`, run:

          $ {command} sample-bpa --project=sample-project --location=us-central1 --backup-plan=sample-backup-plan --resource=sample-resource-uri
        """,
  }

  @staticmethod
  def Args(parser):
    """Specifies additional command flags.

    Args:
      parser: argparse.Parser: Parser object for command line inputs.
    """
    base.ASYNC_FLAG.AddToParser(parser)
    base.ASYNC_FLAG.SetDefault(parser, True)
    flags.AddCreateBackupPlanAssociationFlags(parser)

  def Run(self, args):
    """Constructs and sends request.

    Args:
      args: argparse.Namespace, An object that contains the values for the
        arguments specified in the .Args() method.

    Returns:
      ProcessHttpResponse of the request made.
    """
    api_version = util.GetApiVersion(self.ReleaseTrack())
    client = BackupPlanAssociationsClient(api_version=api_version)
    is_async = args.async_

    backup_plan_association = args.CONCEPTS.backup_plan_association.Parse()
    backup_plan = args.CONCEPTS.backup_plan.Parse()
    resource = args.resource
    resource_type = args.resource_type

    try:
      operation = client.Create(
          backup_plan_association, backup_plan, resource, resource_type
      )
      if resource_type.endswith('Disk'):
        log.warning(
            'In the event of a source region outage, backups in multi-region'
            ' vault remain accessible, but the restore will fail if your disk'
            ' is protected with a regional CMEK.'
        )
    except apitools_exceptions.HttpError as e:
      raise exceptions.HttpException(e, util.HTTP_ERROR_FORMAT)
    if is_async:
      log.CreatedResource(
          backup_plan_association.RelativeName(),
          kind='backup plan association',
          is_async=True,
          details=util.ASYNC_OPERATION_MESSAGE.format(operation.name),
      )
      return operation

    resource = client.WaitForOperation(
        operation_ref=client.GetOperationRef(operation),
        message=(
            'Creating backup plan association [{}]. (This operation could'
            ' take up to 2 minutes.)'.format(
                backup_plan_association.RelativeName()
            )
        ),
    )
    log.CreatedResource(
        backup_plan_association.RelativeName(), kind='backup plan association'
    )

    return resource



"""Generated client library for gkerecommender version v1alpha1."""
# NOTE: This file is autogenerated and should not be edited by hand.

from apitools.base.py import base_api
from googlecloudsdk.generated_clients.apis.gkerecommender.v1alpha1 import gkerecommender_v1alpha1_messages as messages


class GkerecommenderV1alpha1(base_api.BaseApiClient):
  """Generated client library for service gkerecommender version v1alpha1."""

  MESSAGES_MODULE = messages
  BASE_URL = 'https://gkerecommender.googleapis.com/'
  MTLS_BASE_URL = 'https://gkerecommender.mtls.googleapis.com/'

  _PACKAGE = 'gkerecommender'
  _SCOPES = ['https://www.googleapis.com/auth/cloud-platform']
  _VERSION = 'v1alpha1'
  _CLIENT_ID = 'CLIENT_ID'
  _CLIENT_SECRET = 'CLIENT_SECRET'
  _USER_AGENT = 'google-cloud-sdk'
  _CLIENT_CLASS_NAME = 'GkerecommenderV1alpha1'
  _URL_VERSION = 'v1alpha1'
  _API_KEY = None

  def __init__(self, url='', credentials=None,
               get_credentials=True, http=None, model=None,
               log_request=False, log_response=False,
               credentials_args=None, default_global_params=None,
               additional_http_headers=None, response_encoding=None):
    """Create a new gkerecommender handle."""
    url = url or self.BASE_URL
    super(GkerecommenderV1alpha1, self).__init__(
        url, credentials=credentials,
        get_credentials=get_credentials, http=http, model=model,
        log_request=log_request, log_response=log_response,
        credentials_args=credentials_args,
        default_global_params=default_global_params,
        additional_http_headers=additional_http_headers,
        response_encoding=response_encoding)
    self.accelerators = self.AcceleratorsService(self)
    self.modelServers_versions = self.ModelServersVersionsService(self)
    self.modelServers = self.ModelServersService(self)
    self.models = self.ModelsService(self)
    self.modelsAndServers = self.ModelsAndServersService(self)
    self.v1alpha1 = self.V1alpha1Service(self)

  class AcceleratorsService(base_api.BaseApiService):
    """Service class for the accelerators resource."""

    _NAME = 'accelerators'

    def __init__(self, client):
      super(GkerecommenderV1alpha1.AcceleratorsService, self).__init__(client)
      self._upload_configs = {
          }

    def List(self, request, global_params=None):
      r"""List method for the accelerators service.

      Args:
        request: (GkerecommenderAcceleratorsListRequest) input message
        global_params: (StandardQueryParameters, default: None) global arguments
      Returns:
        (ListCompatibleAcceleratorProfilesResponse) The response message.
      """
      config = self.GetMethodConfig('List')
      return self._RunMethod(
          config, request, global_params=global_params)

    List.method_config = lambda: base_api.ApiMethodInfo(
        http_method='GET',
        method_id='gkerecommender.accelerators.list',
        ordered_params=[],
        path_params=[],
        query_params=['modelName', 'modelServerName', 'modelServerVersion', 'performanceRequirements_cost_costPerMillionInputTokens_nanos', 'performanceRequirements_cost_costPerMillionInputTokens_units', 'performanceRequirements_cost_costPerMillionNormalizedOutputTokens_nanos', 'performanceRequirements_cost_costPerMillionNormalizedOutputTokens_units', 'performanceRequirements_cost_outputToInputCostRatio', 'performanceRequirements_cost_pricingModel', 'performanceRequirements_maxNtpotMilliseconds'],
        relative_path='v1alpha1/accelerators',
        request_field='',
        request_type_name='GkerecommenderAcceleratorsListRequest',
        response_type_name='ListCompatibleAcceleratorProfilesResponse',
        supports_download=False,
    )

  class ModelServersVersionsService(base_api.BaseApiService):
    """Service class for the modelServers_versions resource."""

    _NAME = 'modelServers_versions'

    def __init__(self, client):
      super(GkerecommenderV1alpha1.ModelServersVersionsService, self).__init__(client)
      self._upload_configs = {
          }

    def List(self, request, global_params=None):
      r"""List method for the modelServers_versions service.

      Args:
        request: (GkerecommenderModelServersVersionsListRequest) input message
        global_params: (StandardQueryParameters, default: None) global arguments
      Returns:
        (ListModelServerVersionsResponse) The response message.
      """
      config = self.GetMethodConfig('List')
      return self._RunMethod(
          config, request, global_params=global_params)

    List.method_config = lambda: base_api.ApiMethodInfo(
        http_method='GET',
        method_id='gkerecommender.modelServers.versions.list',
        ordered_params=['modelServerName'],
        path_params=['modelServerName'],
        query_params=['modelName'],
        relative_path='v1alpha1/modelServers/{modelServerName}/versions',
        request_field='',
        request_type_name='GkerecommenderModelServersVersionsListRequest',
        response_type_name='ListModelServerVersionsResponse',
        supports_download=False,
    )

  class ModelServersService(base_api.BaseApiService):
    """Service class for the modelServers resource."""

    _NAME = 'modelServers'

    def __init__(self, client):
      super(GkerecommenderV1alpha1.ModelServersService, self).__init__(client)
      self._upload_configs = {
          }

    def List(self, request, global_params=None):
      r"""List method for the modelServers service.

      Args:
        request: (GkerecommenderModelServersListRequest) input message
        global_params: (StandardQueryParameters, default: None) global arguments
      Returns:
        (ListModelServersResponse) The response message.
      """
      config = self.GetMethodConfig('List')
      return self._RunMethod(
          config, request, global_params=global_params)

    List.method_config = lambda: base_api.ApiMethodInfo(
        http_method='GET',
        method_id='gkerecommender.modelServers.list',
        ordered_params=[],
        path_params=[],
        query_params=['modelName'],
        relative_path='v1alpha1/modelServers',
        request_field='',
        request_type_name='GkerecommenderModelServersListRequest',
        response_type_name='ListModelServersResponse',
        supports_download=False,
    )

  class ModelsService(base_api.BaseApiService):
    """Service class for the models resource."""

    _NAME = 'models'

    def __init__(self, client):
      super(GkerecommenderV1alpha1.ModelsService, self).__init__(client)
      self._upload_configs = {
          }

    def List(self, request, global_params=None):
      r"""List method for the models service.

      Args:
        request: (GkerecommenderModelsListRequest) input message
        global_params: (StandardQueryParameters, default: None) global arguments
      Returns:
        (ListModelsResponse) The response message.
      """
      config = self.GetMethodConfig('List')
      return self._RunMethod(
          config, request, global_params=global_params)

    List.method_config = lambda: base_api.ApiMethodInfo(
        http_method='GET',
        method_id='gkerecommender.models.list',
        ordered_params=[],
        path_params=[],
        query_params=[],
        relative_path='v1alpha1/models',
        request_field='',
        request_type_name='GkerecommenderModelsListRequest',
        response_type_name='ListModelsResponse',
        supports_download=False,
    )

  class ModelsAndServersService(base_api.BaseApiService):
    """Service class for the modelsAndServers resource."""

    _NAME = 'modelsAndServers'

    def __init__(self, client):
      super(GkerecommenderV1alpha1.ModelsAndServersService, self).__init__(client)
      self._upload_configs = {
          }

    def List(self, request, global_params=None):
      r"""List method for the modelsAndServers service.

      Args:
        request: (GkerecommenderModelsAndServersListRequest) input message
        global_params: (StandardQueryParameters, default: None) global arguments
      Returns:
        (ListModelAndServerCombinationsResponse) The response message.
      """
      config = self.GetMethodConfig('List')
      return self._RunMethod(
          config, request, global_params=global_params)

    List.method_config = lambda: base_api.ApiMethodInfo(
        http_method='GET',
        method_id='gkerecommender.modelsAndServers.list',
        ordered_params=[],
        path_params=[],
        query_params=['modelName', 'modelServerName', 'modelServerVersion'],
        relative_path='v1alpha1/modelsAndServers',
        request_field='',
        request_type_name='GkerecommenderModelsAndServersListRequest',
        response_type_name='ListModelAndServerCombinationsResponse',
        supports_download=False,
    )

  class V1alpha1Service(base_api.BaseApiService):
    """Service class for the v1alpha1 resource."""

    _NAME = 'v1alpha1'

    def __init__(self, client):
      super(GkerecommenderV1alpha1.V1alpha1Service, self).__init__(client)
      self._upload_configs = {
          }

    def GenerateOptimizationSetManifest(self, request, global_params=None):
      r"""Generates optimized deployment manifests for a given optimization set, GKE version, and other optional input requirements.

      Args:
        request: (GoogleCloudGkerecommenderV1alpha1GenerateOptimizationSetManifestRequest) input message
        global_params: (StandardQueryParameters, default: None) global arguments
      Returns:
        (GoogleCloudGkerecommenderV1alpha1GenerateOptimizationSetManifestResponse) The response message.
      """
      config = self.GetMethodConfig('GenerateOptimizationSetManifest')
      return self._RunMethod(
          config, request, global_params=global_params)

    GenerateOptimizationSetManifest.method_config = lambda: base_api.ApiMethodInfo(
        http_method='POST',
        method_id='gkerecommender.generateOptimizationSetManifest',
        ordered_params=[],
        path_params=[],
        query_params=[],
        relative_path='v1alpha1:generateOptimizationSetManifest',
        request_field='<request>',
        request_type_name='GoogleCloudGkerecommenderV1alpha1GenerateOptimizationSetManifestRequest',
        response_type_name='GoogleCloudGkerecommenderV1alpha1GenerateOptimizationSetManifestResponse',
        supports_download=False,
    )

    def OptimizedManifest(self, request, global_params=None):
      r"""OptimizedManifest method for the v1alpha1 service.

      Args:
        request: (GkerecommenderOptimizedManifestRequest) input message
        global_params: (StandardQueryParameters, default: None) global arguments
      Returns:
        (GenerateOptimizedManifestResponse) The response message.
      """
      config = self.GetMethodConfig('OptimizedManifest')
      return self._RunMethod(
          config, request, global_params=global_params)

    OptimizedManifest.method_config = lambda: base_api.ApiMethodInfo(
        http_method='GET',
        method_id='gkerecommender.optimizedManifest',
        ordered_params=[],
        path_params=[],
        query_params=['acceleratorType', 'kubernetesNamespace', 'modelAndModelServerInfo_modelName', 'modelAndModelServerInfo_modelServerName', 'modelAndModelServerInfo_modelServerVersion', 'storageConfig_modelBucketUri', 'targetNtpotMilliseconds'],
        relative_path='v1alpha1/optimizedManifest',
        request_field='',
        request_type_name='GkerecommenderOptimizedManifestRequest',
        response_type_name='GenerateOptimizedManifestResponse',
        supports_download=False,
    )



#!/usr/bin/env python
"""The BigQuery CLI partition command."""

import datetime
from typing import Optional, cast

from absl import flags

import bq_flags
from clients import client_job
from clients import client_table
from clients import utils as bq_client_utils
from frontend import bigquery_command
from frontend import bq_cached_client
from frontend import utils as frontend_utils
from frontend import utils_flags
from frontend import utils_formatting
from utils import bq_id_utils
from pyglib import stringutil

# These aren't relevant for user-facing docstrings:
# pylint: disable=g-doc-return-or-yield
# pylint: disable=g-doc-args


class Partition(bigquery_command.BigqueryCmd):  # pylint: disable=missing-docstring
  usage = """partition source_prefix destination_table"""
  command = 'partition'

  def __init__(self, name: str, fv: flags.FlagValues):
    super(Partition, self).__init__(name, fv)
    flags.DEFINE_boolean(
        'no_clobber',
        False,
        'Do not overwrite an existing partition.',
        short_name='n',
        flag_values=fv,
    )
    flags.DEFINE_string(
        'time_partitioning_type',
        'DAY',
        'Enables time based partitioning on the table and set the type. The '
        'default value is DAY, which will generate one partition per day. '
        'Other supported values are HOUR, MONTH, and YEAR.',
        flag_values=fv,
    )
    flags.DEFINE_integer(
        'time_partitioning_expiration',
        None,
        'Enables time based partitioning on the table and sets the number of '
        'seconds for which to keep the storage for the partitions in the table.'
        ' The storage in a partition will have an expiration time of its '
        'partition time plus this value.',
        flag_values=fv,
    )
    self._ProcessCommandRc(fv)

  def RunWithArgs(
      self, source_prefix: str, destination_table: str
  ) -> Optional[int]:
    """Copies source tables into partitioned tables.

    Usage:
    bq partition <source_table_prefix> <destination_partitioned_table>

    Copies tables of the format <source_table_prefix><time_unit_suffix> to a
    destination partitioned table, with the <time_unit_suffix> of the source
    tables becoming the partition ID of the destination table partitions. The
    suffix is <YYYYmmdd> by default, <YYYY> if the time_partitioning_type flag
    is set to YEAR, <YYYYmm> if set to MONTH, and <YYYYmmddHH> if set to HOUR.

    If the destination table does not exist, one will be created with
    a schema and that matches the last table that matches the supplied
    prefix.

    Examples:
      bq partition dataset1.sharded_ dataset2.partitioned_table
    """

    client = bq_cached_client.Client.Get()
    formatter = utils_flags.get_formatter_from_flags()

    source_table_prefix = bq_client_utils.GetReference(
        id_fallbacks=client,
        identifier=source_prefix,
        allow_pcnt_identifier_format=True,
    )
    bq_id_utils.typecheck(
        source_table_prefix,
        bq_id_utils.ApiClientHelper.TableReference,
        'Cannot determine table associated with "%s"' % (source_prefix,),
        is_usage_error=True,
    )
    # TODO(b/333595633): Fix typecheck so the response is cast.
    source_table_prefix = cast(
        bq_id_utils.ApiClientHelper.TableReference, source_table_prefix
    )
    destination_table = bq_client_utils.GetReference(
        id_fallbacks=client,
        identifier=destination_table,
        allow_pcnt_identifier_format=True,
    )
    bq_id_utils.typecheck(
        destination_table,
        bq_id_utils.ApiClientHelper.TableReference,
        'Cannot determine table associated with "%s"' % (destination_table,),
        is_usage_error=True,
    )
    # TODO(b/333595633): Fix typecheck so the response is cast.
    destination_table = cast(
        bq_id_utils.ApiClientHelper.TableReference, destination_table
    )

    source_dataset = source_table_prefix.GetDatasetReference()
    source_id_prefix = stringutil.ensure_str(source_table_prefix.tableId)
    source_id_len = len(source_id_prefix)

    job_id_prefix = utils_flags.get_job_id_from_flags()
    if isinstance(job_id_prefix, bq_client_utils.JobIdGenerator):
      job_id_prefix = job_id_prefix.Generate(
          [source_table_prefix, destination_table]
      )

    destination_dataset = destination_table.GetDatasetReference()

    utils_formatting.configure_formatter(
        formatter, bq_id_utils.ApiClientHelper.TableReference
    )
    results = map(
        utils_formatting.format_table_info,
        client_table.list_tables(
            apiclient=client.apiclient,
            reference=source_dataset,
            max_results=1000 * 1000,
        ),
    )

    partition_ids = []
    representative_table = None

    time_format = '%Y%m%d'  # default to format for DAY
    if self.time_partitioning_type == 'HOUR':
      time_format = '%Y%m%d%H'
    elif self.time_partitioning_type == 'MONTH':
      time_format = '%Y%m'
    elif self.time_partitioning_type == 'YEAR':
      time_format = '%Y'

    for result in results:
      table_id = stringutil.ensure_str(result['tableId'])
      if table_id.startswith(source_id_prefix):
        suffix = table_id[source_id_len:]
        try:
          partition_id = datetime.datetime.strptime(suffix, time_format)
          partition_ids.append(partition_id.strftime(time_format))
          representative_table = result
        except ValueError:
          pass

    if not representative_table:
      print('No matching source tables found')
      return

    print(
        'Copying %d source partitions to %s'
        % (len(partition_ids), destination_table)
    )

    # Check to see if we need to create the destination table.
    if not client_table.table_exists(
        apiclient=client.apiclient,
        reference=destination_table,
    ):
      source_table_id = representative_table['tableId']
      source_table_ref = source_dataset.GetTableReference(source_table_id)
      source_table_schema = client_table.get_table_schema(
          apiclient=client.apiclient,
          table_dict=source_table_ref,
      )
      # Get fields in the schema.
      if source_table_schema:
        source_table_schema = source_table_schema['fields']

      time_partitioning = frontend_utils.ParseTimePartitioning(
          self.time_partitioning_type, self.time_partitioning_expiration
      )

      print(
          'Creating table: %s with schema from %s and partition spec %s'
          % (destination_table, source_table_ref, time_partitioning)
      )

      client_table.create_table(
          apiclient=client.apiclient,
          reference=destination_table,
          schema=source_table_schema,
          time_partitioning=time_partitioning,
      )
      print('%s successfully created.' % (destination_table,))

    for partition_id in partition_ids:
      destination_table_id = '%s$%s' % (destination_table.tableId, partition_id)
      source_table_id = '%s%s' % (source_id_prefix, partition_id)
      current_job_id = '%s%s' % (job_id_prefix, partition_id)

      source_table = source_dataset.GetTableReference(source_table_id)
      destination_partition = destination_dataset.GetTableReference(
          destination_table_id
      )

      avoid_copy = False
      if self.no_clobber:
        maybe_destination_partition = client_table.table_exists(
            apiclient=client.apiclient,
            reference=destination_partition,
        )
        avoid_copy = (
            maybe_destination_partition
            and int(maybe_destination_partition['numBytes']) > 0
        )

      if avoid_copy:
        print("Table '%s' already exists, skipping" % (destination_partition,))
      else:
        print('Copying %s to %s' % (source_table, destination_partition))
        kwds = {
            'write_disposition': 'WRITE_TRUNCATE',
            'job_id': current_job_id,
        }
        if bq_flags.LOCATION.value:
          kwds['location'] = bq_flags.LOCATION.value
        job = client_job.CopyTable(
            client, [source_table], destination_partition, **kwds
        )
        if not bq_flags.SYNCHRONOUS_MODE.value:
          self.PrintJobStartInfo(job)
        else:
          print(
              'Successfully copied %s to %s'
              % (source_table, destination_partition)
          )



# -*- coding: utf-8 -*- #
# Copyright 2026 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Command to fetch issues for a database migration conversion workspace."""

import argparse
from typing import Any, Iterator

from googlecloudsdk.api_lib.database_migration import resource_args
from googlecloudsdk.calliope import base
from googlecloudsdk.command_lib.database_migration.conversion_workspaces import command_mixin
from googlecloudsdk.command_lib.database_migration.conversion_workspaces import flags as cw_flags

_DEFAULT_PAGE_SIZE = 100


@base.RegionalEndpointsSupported
@base.Hidden
@base.ReleaseTracks(base.ReleaseTrack.GA)
@base.DefaultUniverseOnly
class FetchIssues(
    command_mixin.ConversionWorkspacesCommandMixin,
    base.ListCommand,
):
  """Fetch issues in a Database Migration conversion workspace."""

  detailed_help = {
      'DESCRIPTION': (
          """
        Fetch issues in a Database Migration conversion workspace.
      """
      ),
      'EXAMPLES': (
          """\
        To fetch the issues in a conversion workspace
        in a project and location `us-central1`, run:

            $ {command} my-conversion-workspace --region=us-central1
      """
      ),
  }

  @staticmethod
  def Args(parser: argparse.ArgumentParser) -> None:
    resource_args.AddConversionWorkspaceResourceArg(parser, 'fetch issues')
    cw_flags.AddAllIssuesFlag(parser)
    base.PAGE_SIZE_FLAG.SetDefault(parser, _DEFAULT_PAGE_SIZE)

    parser.display_info.AddFormat("""
          table(
            entityFullName:label=ENTITY_NAME,
            entityType:label=ENTITY_TYPE,
            severity:label=SEVERITY,
            type:label=TYPE,
            message:label=MESSAGE,
            issueOrigin:label=ORIGIN,
            issueState:label=STATE
          )
        """)

  def Run(
      self,
      args: argparse.Namespace,
  ) -> Iterator[dict[str, Any]]:
    """Fetch issues for a DMS conversion workspace.

    Args:
      args: argparse.Namespace, The arguments that this command was invoked
        with.

    Returns:
      An iterator of issues for the specified conversion workspace and
      arguments.
    """
    conversion_workspace_ref = args.CONCEPTS.conversion_workspace.Parse()
    client = self.GetClient(location=conversion_workspace_ref.locationsId)
    return client.entities.FetchIssues(
        name=conversion_workspace_ref.RelativeName(),
        all_issues=args.all_issues,
        filter_expr=self.ExtractBackendFilter(args),
        page_size=args.GetValue('page_size'),
    )



# -*- coding: utf-8 -*- # Lint as: python3
# Copyright 2026 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Bidirectional conversion between proxy.yaml model and Apigee bundle ZIP.

This module implements the two pure conversion functions called out in the
gcloud Apigee Feature Templates DPD (go/gcloud-apigee-yaml-dpd):

* `proxy_to_bundle(proxy)` -- render a `models.Proxy` into an Apigee bundle
  ZIP (Forward direction).
* `bundle_to_proxy(zip_bytes)` -- parse an Apigee bundle ZIP back into a
  `models.Proxy` (Backward direction).

The implementation is intentionally "throwaway" (per the DPD non-goals): it
ports the conversion algorithm from the Go prototype at
`experimental/users/madhurranjanm/golang-apigee-templater/pkg/converter/`
and will eventually be superseded by the server-side templater service at
`cloud/api_products/apihub/shared/templater/` (go/apihub-templater-api).

Bundle layout (matches Apigee's mgmt API expectation):

    apiproxy/{ProxyName}.xml            # <APIProxy> root descriptor
    apiproxy/proxies/{endpoint}.xml     # one per Proxy.endpoints[i]
    apiproxy/targets/{target}.xml       # one per Proxy.targets[i]
    apiproxy/policies/{policy}.xml      # one per Proxy.policies[i]
    apiproxy/resources/{type}/{name}    # one per Proxy.resources[i]

Policy bodies use the `metadata` / `_text` convention to represent XML
attributes and text content as plain dicts (matching the Go prototype's
`xmlToMap`/`mapToXML`). Round-trip parity is *semantic*, not byte-identical
with Apigee UI emissions.
"""

import io
import posixpath
import re
from typing import Optional, Any
import xml.etree.ElementTree as ET
import zipfile

from googlecloudsdk.command_lib.apigee.aft import models


# ---- Bundle safety bounds (Backward direction) ----

# Total uncompressed bytes across all entries. Mirrors API Hub templater's
# default at cloud/api_products/apihub/shared/templater/converter/zip.go.
_MAX_UNCOMPRESSED_BYTES = 10 * 1024 * 1024  # 10 MiB
_MAX_ENTRIES = 256
_BUNDLE_PREFIX = 'apiproxy/'
_CONFIGURATION_VERSION = '4.0'


class InvalidBundleError(Exception):
  """Raised when a ZIP cannot be interpreted as an Apigee bundle."""


# ============================================================================
# Public API
# ============================================================================


def proxy_to_bundle(proxy: models.Proxy) -> bytes:
  """Renders a Proxy model into an Apigee bundle ZIP.

  Args:
    proxy: A `models.Proxy` instance.

  Returns:
    The bundle ZIP as bytes. The returned bytes are deterministic for a given
    input proxy (entries written in sorted order; dict keys serialized in
    sorted order within each XML element).
  """
  try:
    files = _proxy_to_files(proxy)
  except AttributeError as e:
    raise ValueError(f'{e}; proxy struct: {proxy!r}') from e
  return _pack_files(files)


def bundle_to_proxy(zip_bytes: bytes) -> models.Proxy:
  """Parses an Apigee bundle ZIP into a Proxy model.

  Args:
    zip_bytes: Raw bytes of an Apigee bundle ZIP.

  Returns:
    A `models.Proxy` instance.

  Raises:
    InvalidBundleError: If the input is not a recognizable Apigee bundle,
      exceeds safety bounds, or contains malformed XML.
  """
  files = _unpack_files(zip_bytes)
  return _files_to_proxy(files)


# ============================================================================
# Forward: Proxy -> {path: bytes}
# ============================================================================


def _proxy_to_files(proxy: models.Proxy) -> dict[str, bytes]:
  """Returns a dict mapping bundle ZIP paths to file bytes for `proxy`."""
  files = {}

  # Root descriptor: apiproxy/{name}.xml
  files[_BUNDLE_PREFIX + proxy.name + '.xml'] = _root_xml_bytes(proxy)

  # apiproxy/proxies/{endpoint}.xml
  for endpoint in proxy.endpoints:
    files[_BUNDLE_PREFIX + 'proxies/' + endpoint.name + '.xml'] = (
        _proxy_endpoint_xml_bytes(endpoint)
    )

  # apiproxy/targets/{target}.xml
  for target in proxy.targets:
    files[_BUNDLE_PREFIX + 'targets/' + target.name + '.xml'] = (
        _target_endpoint_xml_bytes(target)
    )

  # apiproxy/policies/{policy}.xml
  for policy in proxy.policies:
    files[_BUNDLE_PREFIX + 'policies/' + policy.name + '.xml'] = (
        _policy_xml_bytes(policy)
    )

  # apiproxy/resources/{type}/{name}
  for resource in proxy.resources:
    path = _BUNDLE_PREFIX + 'resources/' + resource.type + '/' + resource.name
    content = resource.content or ''
    files[path] = content.encode('utf-8')

  return files


def _root_xml_bytes(proxy: models.Proxy) -> bytes:
  """Builds the <APIProxy> root descriptor as bytes."""
  root = ET.Element('APIProxy', attrib={'name': proxy.name})
  _sub_text(root, 'ConfigurationVersion', _CONFIGURATION_VERSION)
  if proxy.description:
    _sub_text(root, 'Description', proxy.description)
  if proxy.display_name:
    _sub_text(root, 'DisplayName', proxy.display_name)

  # Policies list
  policies_el = ET.SubElement(root, 'Policies')
  for policy in proxy.policies:
    _sub_text(policies_el, 'Policy', policy.name)

  # ProxyEndpoints list
  pes_el = ET.SubElement(root, 'ProxyEndpoints')
  for endpoint in proxy.endpoints:
    _sub_text(pes_el, 'ProxyEndpoint', endpoint.name)

  # TargetEndpoints list
  tes_el = ET.SubElement(root, 'TargetEndpoints')
  for target in proxy.targets:
    _sub_text(tes_el, 'TargetEndpoint', target.name)

  # Resources list (URIs of the form `{type}://{name}`).
  resources_el = ET.SubElement(root, 'Resources')
  for resource in proxy.resources:
    _sub_text(resources_el, 'Resource', resource.type + '://' + resource.name)

  return _serialize(root)


def _proxy_endpoint_xml_bytes(endpoint: models.ProxyEndpoint) -> bytes:
  """Builds <ProxyEndpoint> as bytes."""
  root = ET.Element('ProxyEndpoint', attrib={'name': endpoint.name})

  # HTTPProxyConnection / BasePath
  conn_el = ET.SubElement(root, 'HTTPProxyConnection')
  _sub_text(conn_el, 'BasePath', endpoint.base_path or '')

  _add_flows(root, endpoint.flows or [])
  _add_postclientflow(root, endpoint.post_client_flow)

  # RouteRules
  for route in endpoint.routes or []:
    rule_el = ET.SubElement(root, 'RouteRule', attrib={'name': route.name})
    if route.condition is not None:
      _sub_text(rule_el, 'Condition', route.condition)
    if route.target is not None:
      _sub_text(rule_el, 'TargetEndpoint', route.target)

  _add_fault_rules(
      root,
      endpoint.fault_rules or [],
      endpoint.default_fault_rule
  )
  return _serialize(root)


def _target_endpoint_xml_bytes(target: models.Target) -> bytes:
  """Builds <TargetEndpoint> as bytes."""
  root = ET.Element('TargetEndpoint', attrib={'name': target.name})

  built_target = False

  if isinstance(target, models.ProxyTarget):
    # HTTPTargetConnection (URL + optional auth) or LocalTargetConnection
    local_conn = target.local_target_connection
    http_conn = target.http_target_connection
    if local_conn:
      local_el = _dict_to_element('LocalTargetConnection', local_conn)
      root.append(local_el)
      built_target = True
    elif http_conn:
      http_el = _dict_to_element('HTTPTargetConnection', http_conn)
      root.append(http_el)
      built_target = True

    _add_flows(root, target.flows or [])
    _add_fault_rules(
        root,
        target.fault_rules or [],
        target.default_fault_rule,
    )

  if not built_target:
    # Build <HTTPTargetConnection> from simplified Target fields.
    conn_el = ET.SubElement(root, 'HTTPTargetConnection')
    if target.url:
      _sub_text(conn_el, 'URL', target.url)
    if target.auth:
      auth_el = ET.SubElement(conn_el, 'Authentication')
      typed_el = ET.SubElement(auth_el, target.auth)
      if target.scopes:
        scopes_el = ET.SubElement(typed_el, 'Scopes')
        for scope in target.scopes:
          _sub_text(scopes_el, 'Scope', scope)
      if target.aud:
        aud_el = ET.SubElement(typed_el, 'Audience')
        aud_el.text = target.aud

  return _serialize(root)


def _policy_xml_bytes(policy: models.Policy) -> bytes:
  """Builds a policy XML file from `policy.content`.

  The value is rendered using `_dict_to_element` (the inverse of
  `_element_to_dict`).

  Args:
    policy: A `models.Policy` instance. `policy.content` must be a single-key
      dict whose key matches `policy.type`.

  Returns:
    The policy XML as bytes.
  """
  if not isinstance(policy.content, dict) or len(policy.content) != 1:
    raise ValueError(
        'Policy %r content must be a single-key dict; got %r'
        % (policy.name, policy.content)
    )
  ((tag, value),) = policy.content.items()
  if policy.type and tag.lower() != policy.type.lower():
    raise ValueError(
        'Policy %r: type %r does not match content root tag %r'
        % (policy.name, policy.type, tag)
    )
  root = _dict_to_element(tag, value)
  return _serialize(root)


# ---- Flow / FaultRule helpers ----

# Maps a Flow.name to its enclosing XML element name in a ProxyEndpoint /
# TargetEndpoint. Flows with names other than PreFlow / PostFlow live inside
# the generic <Flows> container.
_NAMED_FLOW_TAGS = {
    'PreFlow': 'PreFlow',
    'PostFlow': 'PostFlow',
}


def _add_flows(parent: ET.Element, flows: list[models.Flow]):
  """Appends <PreFlow> / <Flows> / <PostFlow> elements to `parent`.

  Original flow order is preserved by emitting any run of consecutive
  generic (non-PreFlow / non-PostFlow) flows into a single in-position
  `<Flows>` container. PreFlow / PostFlow flows are emitted as top-level
  siblings at their original positions.

  Args:
    parent: The parent XML element.
    flows: The list of flows to add.
  """
  i = 0
  while i < len(flows):
    flow = flows[i]
    tag = _NAMED_FLOW_TAGS.get(flow.name)
    if tag:
      flow_el = ET.SubElement(parent, tag, attrib={'name': flow.name})
      _populate_flow(flow_el, flow)
      i += 1
      continue
    # Collect consecutive generic flows into one <Flows>.
    j = i
    while j < len(flows) and flows[j].name not in _NAMED_FLOW_TAGS:
      j += 1
    flows_el = ET.SubElement(parent, 'Flows')
    for flow in flows[i:j]:
      flow_el = ET.SubElement(flows_el, 'Flow', attrib={'name': flow.name})
      _populate_flow(flow_el, flow)
    i = j


def _add_postclientflow(parent: ET.Element, flow: Optional[models.Flow]):
  """Appends a <PostClientFlow> child for `flow` to `parent` if not None."""
  if flow is None:
    return
  flow_el = ET.SubElement(parent, 'PostClientFlow', attrib={'name': flow.name})
  _populate_flow(flow_el, flow)


def _populate_flow(flow_el: ET.Element, flow: models.Flow):
  """Writes <Request>/<Response>/<Step> children of a flow element.

  If `flow.mode` is unset *and* there are no steps, no wrapper element is
  emitted. This preserves round-trip parity with flows that carry neither
  a mode nor any steps (commonly used as placeholders on target endpoints).

  Args:
    flow_el: The XML element representing the flow.
    flow: The flow model.
  """
  if flow.condition:
    _sub_text(flow_el, 'Condition', flow.condition)
  has_steps = bool(flow.steps)
  if not flow.mode and not has_steps:
    return
  mode = (flow.mode or 'Request').strip()
  if mode not in ('Request', 'Response'):
    mode = 'Request'
  wrapper = ET.SubElement(flow_el, mode)
  for step in flow.steps or []:
    step_el = ET.SubElement(wrapper, 'Step')
    _sub_text(step_el, 'Name', step.name)
    if step.condition is not None:
      _sub_text(step_el, 'Condition', step.condition)


def _add_fault_rules(
    parent: ET.Element,
    fault_rules: list[models.FaultRule],
    default_fault_rule: Optional[models.FaultRule],
):
  """Appends <FaultRules> and <DefaultFaultRule> children to `parent`."""
  if fault_rules:
    frs_el = ET.SubElement(parent, 'FaultRules')
    for rule in fault_rules:
      rule_el = ET.SubElement(frs_el, 'FaultRule', attrib={'name': rule.name})
      _populate_flow(rule_el, rule)
  if default_fault_rule is not None:
    dfr_el = ET.SubElement(
        parent, 'DefaultFaultRule', attrib={'name': default_fault_rule.name}
    )
    if default_fault_rule.always_enforce:
      _sub_text(dfr_el, 'AlwaysEnforce', 'true')
    _populate_flow(dfr_el, default_fault_rule)


# ---- XML serialization helpers ----


def _sub_text(parent: ET.Element, tag: str, text: Optional[str]):
  """Creates a child element with text content."""
  el = ET.SubElement(parent, tag)
  el.text = text if text is not None else ''
  return el


def _serialize(elem: ET.Element) -> bytes:
  """Serializes an Element to bytes with an XML declaration."""
  # `short_empty_elements=False` matches Apigee's preference for explicit
  # close tags on policy bodies.
  return ET.tostring(
      elem, encoding='utf-8', xml_declaration=True, short_empty_elements=False
  )


# ============================================================================
# Backward: {path: bytes} -> Proxy
# ============================================================================


def _files_to_proxy(files: dict[str, bytes]) -> models.Proxy:
  """Reconstructs a Proxy from a `{path: bytes}` file map."""
  # Locate the single root descriptor at apiproxy/*.xml (not in a subdir).
  root_paths = [
      p
      for p in files
      if p.startswith(_BUNDLE_PREFIX)
      and '/' not in p[len(_BUNDLE_PREFIX) :]
      and p.endswith('.xml')
  ]
  if not root_paths:
    raise InvalidBundleError(
        'No root descriptor found (expected exactly one apiproxy/*.xml; got '
        'none).'
    )
  if len(root_paths) > 1:
    root_paths.sort()
    raise InvalidBundleError(f'Multiple root descriptors found: {root_paths}')

  root_path = root_paths[0]
  root_el = _parse_xml(files[root_path], root_path)
  # The bundle has no slot for the authoring-metadata fields (gateway /
  # schemaVersion), so re-stamp the known values here. They must be supplied to
  # the constructor because Proxy.__post_init__ validates them; a
  # post-construction assignment would trip on the name-only construction.
  proxy = models.Proxy(
      name=root_el.get('name') or '',
      gateway='apigee',
      schema_version='1.0.0',
  )

  desc = root_el.find('Description')
  if desc is not None and desc.text:
    proxy.description = desc.text
  dn = root_el.find('DisplayName')
  if dn is not None and dn.text:
    proxy.display_name = dn.text

  # Parse policies (one per apiproxy/policies/*.xml).
  for path in sorted(files):
    rel = path[len(_BUNDLE_PREFIX) :]
    if rel.startswith('policies/') and path.endswith('.xml'):
      proxy.policies.append(_parse_policy(files[path], path))

  # Parse proxy endpoints (one per apiproxy/proxies/*.xml).
  for path in sorted(files):
    rel = path[len(_BUNDLE_PREFIX) :]
    if rel.startswith('proxies/') and path.endswith('.xml'):
      proxy.endpoints.append(_parse_proxy_endpoint(files[path], path))

  # Parse target endpoints (one per apiproxy/targets/*.xml).
  for path in sorted(files):
    rel = path[len(_BUNDLE_PREFIX) :]
    if rel.startswith('targets/') and path.endswith('.xml'):
      proxy.targets.append(_parse_target_endpoint(files[path], path))

  # Resources: apiproxy/resources/{type}/{name}.
  for path in sorted(files):
    rel = path[len(_BUNDLE_PREFIX) :]
    if not rel.startswith('resources/'):
      continue
    parts = rel.split('/', 2)  # ['resources', type, name]
    if len(parts) != 3:
      continue
    _, res_type, res_name = parts
    proxy.resources.append(
        models.Resource(
            name=res_name,
            type=res_type,
            content=files[path].decode('utf-8', errors='replace'),
        )
    )

  return proxy


def _parse_xml(data: bytes, path: str) -> ET.Element:
  try:
    return ET.fromstring(data)
  except ET.ParseError as e:
    raise InvalidBundleError(f'Malformed XML in {path}: {e}') from e


def _parse_policy(data: bytes, path: str) -> models.Policy:
  root = _parse_xml(data, path)
  name_attr = root.get('name') or _basename(path)
  return models.Policy(
      name=name_attr, type=root.tag, content={root.tag: _element_to_dict(root)}
  )


def _basename(path: str) -> str:
  base = posixpath.basename(path)
  return base[:-4] if base.endswith('.xml') else base


def _parse_proxy_endpoint(data: bytes, path: str) -> models.ProxyEndpoint:
  """Parses an apiproxy/proxies/*.xml file into a `models.ProxyEndpoint`."""
  root = _parse_xml(data, path)
  endpoint = models.ProxyEndpoint(name=root.get('name') or _basename(path))

  conn = root.find('HTTPProxyConnection')
  if conn is not None:
    bp = conn.find('BasePath')
    if bp is not None and bp.text is not None:
      endpoint.base_path = bp.text

  endpoint.routes = _parse_route_rules(root)
  endpoint.flows = _parse_flows(root)
  pcf = root.find('PostClientFlow')
  if pcf is not None:
    endpoint.post_client_flow = _parse_named_flow(pcf)
  endpoint.fault_rules, endpoint.default_fault_rule = _parse_fault_rules(root)
  return endpoint


def _parse_target_endpoint(data: bytes, path: str) -> models.ProxyTarget:
  """Parses an apiproxy/targets/*.xml file into a `models.ProxyTarget`."""
  root = _parse_xml(data, path)
  target = models.ProxyTarget(name=root.get('name') or _basename(path))

  local = root.find('LocalTargetConnection')
  if local is not None:
    target.local_target_connection = _element_to_dict(local)
  http = root.find('HTTPTargetConnection')
  if http is not None:
    url_el = http.find('URL')
    if url_el is not None and url_el.text is not None:
      target.url = url_el.text
    auth_el = http.find('Authentication')
    if auth_el is not None:
      # The single typed child of <Authentication> names the auth scheme
      # (e.g. <GoogleAccessToken>, <GoogleIDToken>).
      typed_children = list(auth_el)
      if typed_children:
        scheme_el = typed_children[0]
        target.auth = scheme_el.tag
        scopes_el = scheme_el.find('Scopes')
        if scopes_el is not None:
          target.scopes = [
              s.text for s in scopes_el.findall('Scope') if s.text is not None
          ]
        aud_el = scheme_el.find('Audience')
        if aud_el is not None and aud_el.text is not None:
          target.aud = aud_el.text

  target.flows = _parse_flows(root)
  target.fault_rules, target.default_fault_rule = _parse_fault_rules(root)
  return target


def _parse_route_rules(root: ET.Element) -> list[models.Route]:
  """Returns a list of `models.Route` from `<RouteRule>` children of `root`."""
  rules = []
  for rule_el in root.findall('RouteRule'):
    route = models.Route(name=rule_el.get('name') or '')
    cond = rule_el.find('Condition')
    if cond is not None and cond.text is not None:
      route.condition = cond.text
    tgt = rule_el.find('TargetEndpoint')
    if tgt is not None and tgt.text is not None:
      route.target = tgt.text
    rules.append(route)
  return rules


def _parse_named_flow(flow_el: ET.Element) -> models.Flow:
  """Parses a <PreFlow>/<PostFlow>/<Flow>/<PostClientFlow> element."""
  name_attr = flow_el.get('name')
  # Treat a missing `name=` attribute as the tag name; treat `name=""` as
  # the empty-string flow name that the AFT YAML schema allows.
  name = flow_el.tag if name_attr is None else name_attr
  mode = None
  steps = []
  for child in flow_el:
    if child.tag in ('Request', 'Response'):
      mode = child.tag
      for step_el in child.findall('Step'):
        step = models.Step()
        name_el = step_el.find('Name')
        if name_el is not None and name_el.text is not None:
          step.name = name_el.text
        cond = step_el.find('Condition')
        if cond is not None and cond.text is not None:
          step.condition = cond.text
        steps.append(step)
  flow = models.Flow(name=name, mode=mode, steps=steps)
  cond_el = flow_el.find('Condition')
  if cond_el is not None and cond_el.text is not None:
    flow.condition = cond_el.text
  return flow


def _parse_flows(root: ET.Element) -> list[models.Flow]:
  """Returns the flow list in original (document) order.

  Walks `root`'s direct children: <PreFlow> / <PostFlow> become standalone
  Flow entries; each <Flows> container contributes its inner <Flow>
  entries in document order.

  Args:
    root: The XML element representing the flow.

  Returns:
    A list of `models.Flow` objects.
  """
  flows = []
  for child in root:
    if child.tag in ('PreFlow', 'PostFlow'):
      flows.append(_parse_named_flow(child))
    elif child.tag == 'Flows':
      for flow_el in child.findall('Flow'):
        flows.append(_parse_named_flow(flow_el))
  return flows


def _parse_fault_rules(
    root: ET.Element,
) -> tuple[list[models.FaultRule], Optional[models.FaultRule]]:
  """Returns (fault_rules, default_fault_rule) parsed from `root`'s children."""
  fault_rules = []
  frs_container = root.find('FaultRules')
  if frs_container is not None:
    for el in frs_container.findall('FaultRule'):
      fault_rules.append(_parse_named_flow(el))
  default_rule = None
  dfr_el = root.find('DefaultFaultRule')
  if dfr_el is not None:
    base_flow = _parse_named_flow(dfr_el)
    default_rule = models.FaultRule(
        name=base_flow.name,
        mode=base_flow.mode,
        condition=base_flow.condition,
        steps=base_flow.steps,
        always_enforce=_text_to_bool(dfr_el.find('AlwaysEnforce')),
    )
  return fault_rules, default_rule


def _text_to_bool(el: Optional[ET.Element]) -> bool:
  if el is None or el.text is None:
    return False
  return el.text.strip().lower() == 'true'


# ============================================================================
# XML <-> dict (`metadata` / `_text` convention)
# ============================================================================


def _element_to_dict(elem: ET.Element) -> dict[str, Any]:
  """Converts an ElementTree element to a JSON-ish dict.

  The convention (ported from the Go prototype's `xmlToMap`):

  * Element attributes go under the `metadata` key.
  * Element text (non-empty after stripping pure-whitespace) goes under `_text`.
  * Child elements are nested under their tag name.
  * Repeated child tags become a list of dicts under that single key.

  Example:
    <Foo attr="x">hi<Bar/><Bar/></Foo>
    -> {"metadata": {"attr": "x"}, "_text": "hi", "Bar": [{}, {}]}

  Args:
    elem: The ElementTree element to convert.

  Returns:
    A dict representation of the element.
  """
  result = {}
  if elem.attrib:
    # Sort attribute keys for deterministic serialization.
    result['metadata'] = {k: elem.attrib[k] for k in sorted(elem.attrib)}
  text = (elem.text or '').strip()
  if text:
    result['_text'] = elem.text  # preserve original whitespace
  for child in elem:
    child_dict = _element_to_dict(child)
    if child.tag in result:
      existing = result[child.tag]
      if isinstance(existing, list):
        existing.append(child_dict)
      else:
        result[child.tag] = [existing, child_dict]
    elif tuple(child_dict.keys()) == ('_text',):
      # _text sometimes can't be avoided, e.g. if the element has attributes...
      # but if the child is just text, store it directly under the parent key to
      # make the YAML cleaner.
      result[child.tag] = child_dict['_text']
    else:
      result[child.tag] = child_dict
  return result


def _dict_to_element(tag: str, value: Any) -> ET.Element:
  """Inverse of `_element_to_dict`. Returns an Element tagged `tag`."""
  elem = ET.Element(tag)
  if value is None:
    return elem
  if not isinstance(value, dict):
    # Bare scalar -> element text.
    elem.text = _as_text(value)
    return elem

  for key in sorted(value):
    val = value[key]
    if key == 'metadata':
      if not isinstance(val, dict):
        raise ValueError(
            f'metadata for <{tag}> must be a dict; got {type(val)}'
        )
      for ak in sorted(val):
        elem.set(ak, _as_text(val[ak]))
    elif key == '_text':
      elem.text = _as_text(val)
    else:
      _append_child(elem, key, val)
  return elem


def _append_child(parent: ET.Element, tag: str, value: Any):
  if isinstance(value, list):
    for item in value:
      parent.append(_dict_to_element(tag, item))
  else:
    parent.append(_dict_to_element(tag, value))


def _as_text(value: Any) -> str:
  """Coerces a YAML-loaded scalar to an XML-safe text string."""
  if value is None:
    return ''
  if isinstance(value, bool):
    return 'true' if value else 'false'
  return str(value)


# ============================================================================
# ZIP packing / unpacking
# ============================================================================


def _pack_files(files):
  """Packs a `{path: bytes}` dict into a deterministic ZIP."""
  buf = io.BytesIO()
  with zipfile.ZipFile(buf, mode='w', compression=zipfile.ZIP_DEFLATED) as zf:
    for path in sorted(files):
      info = zipfile.ZipInfo(filename=path)
      # Pin date_time so the output is deterministic across runs.
      info.date_time = (1980, 1, 1, 0, 0, 0)
      info.compress_type = zipfile.ZIP_DEFLATED
      zf.writestr(info, files[path])
  return buf.getvalue()


# Reject path entries containing traversal segments or absolute paths.
_PATH_TRAVERSAL_RE = re.compile(r'(^/|(^|/)\.\.(/|$))')


def _unpack_files(zip_bytes: bytes) -> dict[str, bytes]:
  """Reads a ZIP into a `{path: bytes}` dict, applying safety bounds.

  Args:
    zip_bytes: The bytes of the ZIP file.

  Returns:
    A dict mapping normalized path names to content bytes.

  Raises:
    InvalidBundleError: On bad ZIP, traversal attempt, oversized content,
      entry-count overflow, or entries outside `apiproxy/`.
  """
  try:
    zf = zipfile.ZipFile(io.BytesIO(zip_bytes), mode='r')
  except zipfile.BadZipFile as e:
    raise InvalidBundleError(f'Not a valid ZIP: {e}') from e

  infos = zf.infolist()
  if len(infos) > _MAX_ENTRIES:
    raise InvalidBundleError(
        f'Bundle has {len(infos)} entries; max is {_MAX_ENTRIES}.'
    )

  total = 0
  files = {}
  for info in infos:
    name = info.filename
    if name.endswith('/'):
      # Directory entries -- ignore.
      continue
    if _PATH_TRAVERSAL_RE.search(name):
      raise InvalidBundleError(f'Path traversal entry rejected: {name!r}')
    if not name.startswith(_BUNDLE_PREFIX):
      raise InvalidBundleError(f'Entry {name!r} is outside {_BUNDLE_PREFIX}/')
    if info.file_size > _MAX_UNCOMPRESSED_BYTES:
      raise InvalidBundleError(
          f'Entry {name!r} declares uncompressed size {info.file_size} > cap'
          f' {_MAX_UNCOMPRESSED_BYTES}.'
      )
    total += info.file_size
    if total > _MAX_UNCOMPRESSED_BYTES:
      raise InvalidBundleError(
          f'Total declared uncompressed size > cap {_MAX_UNCOMPRESSED_BYTES}.'
      )
    if name in files:
      raise InvalidBundleError(f'Duplicate entry: {name!r}')
    files[name] = zf.read(info)

  return files



# -*- coding: utf-8 -*- #
# Copyright 2020 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Command that restarts web server for an environment."""


from googlecloudsdk.api_lib.composer import environments_util as environments_api_util
from googlecloudsdk.api_lib.composer import operations_util as operations_api_util
from googlecloudsdk.calliope import base
from googlecloudsdk.command_lib.composer import resource_args
from googlecloudsdk.command_lib.composer import util as command_util
from googlecloudsdk.core import log

DETAILED_HELP = {
    'EXAMPLES':
        """\
          To restart the Cloud Composer web server in an environment named
          ``env-1'', run:

            $ {command} env-1
        """
}


@base.ReleaseTracks(
    base.ReleaseTrack.ALPHA, base.ReleaseTrack.BETA, base.ReleaseTrack.GA
)
# TODO(b/371178112): Reconsider this annotation once the command is ready for
# other universes.
@base.DefaultUniverseOnly
class RestartWebServer(base.Command):
  """Restart web server for a Cloud Composer environment."""

  detailed_help = DETAILED_HELP

  @staticmethod
  def Args(parser):
    resource_args.AddEnvironmentResourceArg(parser, 'to restart web server for')
    base.ASYNC_FLAG.AddToParser(parser)

  def Run(self, args):
    env_resource = args.CONCEPTS.environment.Parse()
    operation = environments_api_util.RestartWebServer(
        env_resource, release_track=self.ReleaseTrack())
    if args.async_:
      return self._AsynchronousExecution(env_resource, operation)
    else:
      return self._SynchronousExecution(env_resource, operation)

  def _AsynchronousExecution(self, env_resource, operation):
    details = 'with operation [{0}]'.format(operation.name)
    log.UpdatedResource(
        env_resource.RelativeName(),
        kind='environment',
        is_async=True,
        details=details)
    return operation

  def _SynchronousExecution(self, env_resource, operation):
    try:
      operations_api_util.WaitForOperation(
          operation,
          'Waiting for [{}] to be updated with [{}]'.format(
              env_resource.RelativeName(), operation.name),
          release_track=self.ReleaseTrack())
    except command_util.Error as e:
      raise command_util.Error(
          'Error restarting web server [{}]: {}'.format(
              env_resource.RelativeName(), str(e)
          )
      )



# -*- coding: utf-8 -*- #
# Copyright 2017 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""endpoints deploy command."""


import os
from typing import Final
import urllib.parse

from googlecloudsdk.api_lib.endpoints import config_reporter
from googlecloudsdk.api_lib.endpoints import exceptions
from googlecloudsdk.api_lib.endpoints import services_util
from googlecloudsdk.api_lib.services import enable_api
from googlecloudsdk.api_lib.services import exceptions as services_exceptions
from googlecloudsdk.calliope import base
from googlecloudsdk.calliope import exceptions as calliope_exceptions
from googlecloudsdk.core import log
from googlecloudsdk.core import properties
from googlecloudsdk.core.console import console_io
from googlecloudsdk.core.util import http_encoding


ADVICE_STRING = (
    'Advice found for changes in the new service config. If this '
    'is a --validate-only run, the config push would have failed. '
    'See the outputted report for failure reason(s). If this is '
    'not a --validate-only run and you would like to ignore these '
    'warnings, rerun the command with --force. NOTE: setting this '
    'flag will ignore all change advice. For production systems, '
    'best practice is to set this for a single execution only '
    'after manually reviewing all changes with advice.'
)

FORCE_ADVICE_STRING = (
    'Advice found for changes in the new service config, '
    'but proceeding anyway because --force is set...'
)

VALIDATE_NEW_PROMPT = (
    'The service {service_name} must exist in order to '
    'validate the configuration. Do you want to create the '
    'service in project {project_id}?'
)
VALIDATE_NEW_ERROR = (
    'The service {service_name} must exist in order to '
    'validate the configuration. To create the service in '
    'project {project_id}, rerun the command without the '
    '--validate-only flag.'
)

NUM_ADVICE_TO_PRINT = 3

SWAGGER_VERSION: Final[str] = '2.x'
OPENAPI_V3_VERSION: Final[str] = '3.x'


def _CommonArgs(parser):
  """Add common arguments for this command to the given parser."""
  parser.add_argument(
      'service_config_file',
      nargs='+',
      help=(
          'The service configuration file (or files) containing the API '
          'specification to upload. Proto Descriptors, Open API (Swagger) '
          'specifications, and Google Service Configuration files in JSON '
          'and YAML formats are acceptable.'
      ),
  )
  base.ASYNC_FLAG.AddToParser(parser)


def GenerateManagementUrl(service):
  """Generate a service management url for this service."""
  # It is actually possible to deploy a service configuration for a service
  # which is not in the current project, as long as you have appropriate
  # permissions. Because of this, we need to explicitly query for the service's
  # project.
  messages = services_util.GetMessagesModule()
  client = services_util.GetClientInstance()
  get_request = messages.ServicemanagementServicesGetRequest(
      serviceName=service,
  )
  response = client.services.Get(get_request)
  project = response.producerProjectId

  return (
      'https://console.cloud.google.com/endpoints/api/'
      '{service}/overview?project={project}'.format(
          service=urllib.parse.quote(service),
          project=urllib.parse.quote(project),
      )
  )


class _BaseDeploy(object):
  """Create deploy base class for all release tracks."""

  @staticmethod
  def Args(parser):
    """Args is called by calliope to gather arguments for this command.

    Args:
      parser: An argparse parser that you can use to add arguments that go on
        the command line after this command. Positional arguments are allowed.
    """
    _CommonArgs(parser)
    parser.add_argument(
        '--validate-only',
        action='store_true',
        help=(
            'If included, the command validates the service configuration(s), '
            'but does not deploy them. The service must exist in order to '
            'validate the configuration(s).'
        ),
    )

    parser.add_argument(
        '--force',
        '-f',
        action='store_true',
        default=False,
        help=(
            'Force the deployment even if any hazardous changes to the '
            'service configuration are detected.'
        ),
    )

  def MakeConfigFileMessage(
      self,
      file_contents: str,
      filename: str,
      file_type,
  ):
    """Constructs a ConfigFile message from a config file.

    Args:
      file_contents: The contents of the config file.
      filename: The full path to the config file.
      file_type: FileTypeValueValuesEnum describing the type of config file.

    Returns:
      The constructed ConfigFile message.
    """

    messages = services_util.GetMessagesModule()

    file_types = messages.ConfigFile.FileTypeValueValuesEnum
    if file_type != file_types.FILE_DESCRIPTOR_SET_PROTO:
      # File is human-readable text, not binary; needs to be encoded.
      file_contents = http_encoding.Encode(file_contents)
    return messages.ConfigFile(
        fileContents=file_contents,
        filePath=os.path.basename(filename),
        fileType=file_type,
    )

  def ShowConfigReport(self, service, service_config_id, log_func=log.warning):
    """Run and display results (if any) from the Push Advisor.

    Args:
      service: The name of the service for which to compare configs.
      service_config_id: The new config ID to compare against the active config.
      log_func: The function to which to pass advisory messages (default:
        log.warning).

    Returns:
      The number of advisory messages returned by the Push Advisor.
    """
    num_changes_with_advice = 0

    reporter = config_reporter.ConfigReporter(service)

    # Set the new config as the recently generated service config ID
    reporter.new_config.SetConfigId(service_config_id)

    # We always want to compare agaisnt the active config, so use default here
    reporter.old_config.SetConfigUseDefaultId()

    change_report = reporter.RunReport()
    if not change_report or not change_report.configChanges:
      return 0

    changes = change_report.configChanges

    for change in changes:
      if change.advices:
        if num_changes_with_advice < NUM_ADVICE_TO_PRINT:
          log_func(
              '%s\n', services_util.PushAdvisorConfigChangeToString(change)
          )
        num_changes_with_advice += 1

    if num_changes_with_advice > NUM_ADVICE_TO_PRINT:
      log_func(
          '%s total changes with advice found, check config report file '
          'for full list.\n',
          num_changes_with_advice,
      )

    return num_changes_with_advice

  def CheckPushAdvisor(self, unused_force=False):
    """Run the Push Advisor and return whether the command should abort.

    Args:
      unused_force: bool, unused in the default implementation.

    Returns:
      True if the deployment should be aborted due to warnings, otherwise
      False if it's safe to continue.
    """
    # Child classes must override this; otherwise, we'll always return False
    return False

  def AttemptToEnableService(self, service_name, is_async):
    """Attempt to enable a service. If lacking permission, log a warning.

    Args:
      service_name: The service to enable.
      is_async: If true, return immediately instead of waiting for the operation
        to finish.
    """
    project_id = properties.VALUES.core.project.Get(required=True)
    try:
      # Enable the produced service.
      enable_api.EnableService(project_id, service_name, is_async)
      # The above command will print a message to the human user, but it needs a
      # newline when the command is successful.
      log.status.Print('\n')
    except services_exceptions.EnableServiceException:
      log.warning(
          (
              'Attempted to enable service [{0}] on project [{1}], but '
              'did not have required permissions. Please ensure this '
              'service is enabled before using your Endpoints '
              'service.\n'
          ).format(service_name, project_id)
      )

  def Run(self, args):
    """Run 'endpoints services deploy'.

    Args:
      args: argparse.Namespace, The arguments that this command was invoked
        with.

    Returns:
      The response from the Update API call.

    Raises:
      BadFileExceptionn: if the provided service configuration files are
          invalid or cannot be read.
    """
    messages = services_util.GetMessagesModule()
    client = services_util.GetClientInstance()

    file_types = messages.ConfigFile.FileTypeValueValuesEnum
    self.service_name = self.service_version = config_contents = None
    config_files = []

    self.validate_only = args.validate_only

    # TODO(b/77867100): remove .proto support and deprecation warning.
    give_proto_deprecate_warning = False

    # If we're not doing a validate-only run, we don't want to output the
    # resource directly unless the user specifically requests it using the
    # --format flag. The Epilog will show useful information after deployment
    # is complete.
    if not self.validate_only and not args.IsSpecified('format'):
      args.format = 'none'

    oas_version = None
    for i, service_config_file in enumerate(args.service_config_file):
      config_contents = services_util.ReadServiceConfigFile(service_config_file)

      if services_util.FilenameMatchesExtension(
          service_config_file, ['.json', '.yaml', '.yml']
      ):
        # Try to load the file as JSON. If that fails, try YAML.
        service_config_dict = services_util.LoadJsonOrYaml(config_contents)
        if not service_config_dict:
          raise calliope_exceptions.BadFileException(
              'Could not read JSON or YAML from service config file '
              '[{0}].'.format(service_config_file)
          )

        if 'swagger' in service_config_dict:
          # Only one OpenAPI version is supported per deployment.
          if oas_version and oas_version != SWAGGER_VERSION:
            raise calliope_exceptions.BadFileException(
                (
                    'Malformed input. Found OpenAPI service config in file'
                    ' [{}], but a different OpenAPI version was'
                    ' specified in another file. Only one OpenAPI version is'
                    ' supported per deployment.'
                ).format(service_config_file)
            )
          if not oas_version:
            oas_version = SWAGGER_VERSION

          if 'host' not in service_config_dict:
            raise calliope_exceptions.BadFileException(
                (
                    'Malformed input. Found Swagger service config in file'
                    ' [{}], but no host was specified. Add a host specification'
                    ' to the config file.'
                ).format(service_config_file)
            )
          if not self.service_name and service_config_dict.get('host'):
            self.service_name = service_config_dict.get('host')

          # Always use YAML for Open API because JSON is a subset of YAML.
          config_files.append(
              self.MakeConfigFileMessage(
                  config_contents, service_config_file, file_types.OPEN_API_YAML
              )
          )
        elif 'openapi' in service_config_dict:
          # Only one OpenAPI version is supported per deployment.
          if oas_version and oas_version != OPENAPI_V3_VERSION:
            raise calliope_exceptions.BadFileException(
                (
                    'Malformed input. Found OpenAPI service config in file'
                    ' [{}], but a different OpenAPI version was'
                    ' specified in another file. Only one OpenAPI version is'
                    ' supported per deployment.'
                ).format(service_config_file)
            )
          if not oas_version:
            oas_version = OPENAPI_V3_VERSION

          # For Endpoints, exactly only one server may be configured with the
          # "x-google-endpoint" extension per file and one must be configured
          # across all provided files. The server's hostname will be used as the
          # service name. If multiple oas files are provided, the service name
          # must be the same across all files.
          found_x_google_endpoint = False
          x_google_endpoint_service_name = None
          for s in service_config_dict.get('servers') or []:
            if 'x-google-endpoint' in s:
              if found_x_google_endpoint:
                raise calliope_exceptions.BadFileException(
                    (
                        'Malformed input. Found OpenAPI service config in file'
                        ' [{}], but multiple servers were specified with the'
                        ' "x-google-endpoint" extension. At most one server'
                        ' can have the "x-google-endpoint" extension per file.'
                    ).format(service_config_file)
                )
              if 'url' not in s:
                raise calliope_exceptions.BadFileException(
                    (
                        'Malformed input. Found OpenAPI service config in file'
                        ' [{}], but no url was specified for the server with'
                        ' the "x-google-endpoint" extension. Add a url'
                        ' specification to the server with the'
                        ' "x-google-endpoint" extension.'
                    ).format(service_config_file)
                )
              x_google_endpoint_service_name = urllib.parse.urlparse(
                  s.get('url')
              ).hostname
              found_x_google_endpoint = True

          # If this is the first file with the "x-google-endpoint" extension,
          # use the server URL's hostname as the service name.
          if self.service_name is None and x_google_endpoint_service_name:
            self.service_name = x_google_endpoint_service_name

          # If this is the last (or only) file and no server with the
          # "x-google-endpoint" extension was found, raise an error.
          if (
              i == len(args.service_config_file) - 1
              and self.service_name is None
          ):
            raise calliope_exceptions.BadFileException(
                'Malformed input. No servers were specified with the'
                ' "x-google-endpoint" extension in the provided OpenAPI'
                ' file(s). Add a servers specification to at least one config'
                ' file with exactly one server configured with the'
                ' "x-google-endpoint" extension.'
            )

          # If multiple files are provided, ensure that all files with the
          # "x-google-endpoint" extension define the same server URL. This
          # ensures that the same service name is used for all files.
          if (
              self.service_name
              and x_google_endpoint_service_name
              and self.service_name != x_google_endpoint_service_name
          ):
            raise calliope_exceptions.BadFileException((
                'Ambiguous input. Multiple OpenAPI files were provided with'
                ' different server URLs containing the "x-google-endpoint"'
                ' extension. When using multiple OpenAPI files, ensure that all'
                ' files with the "x-google-endpoint" extension define the same'
                ' server URL with at most one server with the'
                ' "x-google-endpoint" extension defined per file. Additionally,'
                ' at least one server across all provided files must have the'
                ' "x-google-endpoint" extension.'
            ))

          # Always use YAML for Open API because JSON is a subset of YAML.
          config_files.append(
              self.MakeConfigFileMessage(
                  config_contents, service_config_file, file_types.OPEN_API_YAML
              )
          )
        elif service_config_dict.get('type') == 'google.api.Service':
          if not self.service_name and service_config_dict.get('name'):
            self.service_name = service_config_dict.get('name')

          config_files.append(
              self.MakeConfigFileMessage(
                  config_contents,
                  service_config_file,
                  file_types.SERVICE_CONFIG_YAML,
              )
          )
        elif 'name' in service_config_dict:
          # This is a special case. If we have been provided a Google Service
          # Configuration file which has a service 'name' field, but no 'type'
          # field, we have to assume that this is a normalized service config,
          # and can be uploaded via the CreateServiceConfig API. Therefore,
          # we can short circute the process here.
          if len(args.service_config_file) > 1:
            raise calliope_exceptions.BadFileException(
                (
                    'Ambiguous input. Found normalized service configuration in'
                    ' file [{0}], but received multiple input files. To upload'
                    ' normalized service config, please provide it separately'
                    ' from other input files to avoid ambiguity.'
                ).format(service_config_file)
            )

          # If this is a validate-only run, abort now, since this is not
          # supported in the ServiceConfigs.Create API
          if self.validate_only:
            raise exceptions.InvalidFlagError(
                'The --validate-only flag is not supported when using '
                'normalized service configs as input.'
            )

          self.service_name = service_config_dict.get('name')
          config_files = []
          break
        else:
          raise calliope_exceptions.BadFileException(
              (
                  'Unable to parse Open API, or Google Service Configuration '
                  'specification from {0}'
              ).format(service_config_file)
          )

      elif services_util.IsProtoDescriptor(service_config_file):
        config_files.append(
            self.MakeConfigFileMessage(
                config_contents,
                service_config_file,
                file_types.FILE_DESCRIPTOR_SET_PROTO,
            )
        )
      elif services_util.IsRawProto(service_config_file):
        give_proto_deprecate_warning = True
        config_files.append(
            self.MakeConfigFileMessage(
                config_contents, service_config_file, file_types.PROTO_FILE
            )
        )
      else:
        raise calliope_exceptions.BadFileException(
            (
                'Could not determine the content type of file [{0}]. Supported '
                'extensions are .json .yaml .yml .pb and .descriptor'
            ).format(service_config_file)
        )

    if give_proto_deprecate_warning:
      log.warning(
          'Support for uploading uncompiled .proto files is deprecated and '
          'will soon be removed. Use compiled descriptor sets (.pb) instead.\n'
      )

    # Check if we need to create the service.
    was_service_created = False
    if not services_util.DoesServiceExist(self.service_name):
      project_id = properties.VALUES.core.project.Get(required=True)
      # Deploying, even with validate-only, cannot succeed without the service
      # being created
      if self.validate_only:
        if not console_io.CanPrompt():
          raise exceptions.InvalidConditionError(
              VALIDATE_NEW_ERROR.format(
                  service_name=self.service_name, project_id=project_id
              )
          )
        if not console_io.PromptContinue(
            VALIDATE_NEW_PROMPT.format(
                service_name=self.service_name, project_id=project_id
            )
        ):
          return None
      services_util.CreateService(self.service_name, project_id)
      was_service_created = True

    if config_files:
      push_config_result = services_util.PushMultipleServiceConfigFiles(
          self.service_name,
          config_files,
          args.async_,
          validate_only=self.validate_only,
      )
      self.service_config_id = (
          services_util.GetServiceConfigIdFromSubmitConfigSourceResponse(
              push_config_result
          )
      )
    else:
      push_config_result = services_util.PushNormalizedGoogleServiceConfig(
          self.service_name,
          properties.VALUES.core.project.Get(required=True),
          services_util.LoadJsonOrYaml(config_contents),
      )
      self.service_config_id = push_config_result.id

    if not self.service_config_id:
      raise exceptions.InvalidConditionError(
          'Failed to retrieve Service Configuration Id.'
      )

    # Run the Push Advisor to see if we need to warn the user of any
    # potentially hazardous changes to the service configuration.
    if self.CheckPushAdvisor(args.force):
      return None

    # Create a Rollout for the new service configuration
    if not self.validate_only:
      percentages = messages.TrafficPercentStrategy.PercentagesValue()
      percentages.additionalProperties.append((
          messages.TrafficPercentStrategy.PercentagesValue.AdditionalProperty(
              key=self.service_config_id, value=100.0
          )
      ))
      traffic_percent_strategy = messages.TrafficPercentStrategy(
          percentages=percentages
      )
      rollout = messages.Rollout(
          serviceName=self.service_name,
          trafficPercentStrategy=traffic_percent_strategy,
      )
      rollout_create = messages.ServicemanagementServicesRolloutsCreateRequest(
          rollout=rollout,
          serviceName=self.service_name,
      )
      rollout_operation = client.services_rollouts.Create(rollout_create)
      services_util.ProcessOperationResult(rollout_operation, args.async_)

      if was_service_created:
        self.AttemptToEnableService(self.service_name, args.async_)

    return push_config_result

  def Epilog(self, resources_were_displayed):
    # Print this to screen not to the log because the output is needed by the
    # human user. Only print this when not doing a validate-only run.
    if resources_were_displayed and not self.validate_only:
      log.status.Print(
          ('Service Configuration [{0}] uploaded for service [{1}]\n').format(
              self.service_config_id, self.service_name
          )
      )

      management_url = GenerateManagementUrl(self.service_name)
      log.status.Print('To manage your API, go to: ' + management_url)


@base.ReleaseTracks(base.ReleaseTrack.GA)
@base.DefaultUniverseOnly
class Deploy(_BaseDeploy, base.Command):
  # pylint: disable=line-too-long
  """Deploys a service configuration for the given service name.

  This command is used to deploy a service configuration for a service
  to Google Service Management. As input, it takes one or more paths
  to service configurations that should be uploaded. These configuration
  files can be Proto Descriptors, Open API (Swagger) specifications,
  or Google Service Configuration files in JSON or YAML formats.

  If a service name is present in multiple configuration files (given
  in the `host` field in OpenAPI specifications or the `name` field in
  Google Service Configuration files), the first one will take precedence.

  This command will block until deployment is complete unless the
  `--async` flag is passed.

  ## EXAMPLES
  To deploy a single Open API service configuration, run:

    $ {command} ~/my_app/openapi.json

  To run the deployment asynchronously (non-blocking), run:

    $ {command} ~/my_app/openapi.json --async

  To deploy a service config with a Proto, run:

    $ {command} ~/my_app/service-config.yaml ~/my_app/service-protos.pb
  """
  # pylint: enable=line-too-long


# TODO(b/65455903) Once PushAdvisor is ready to be included by default,
# merge this back into _BaseDeploy and rename it to Deploy.
@base.ReleaseTracks(base.ReleaseTrack.ALPHA, base.ReleaseTrack.BETA)
@base.DefaultUniverseOnly
class DeployBetaAlpha(_BaseDeploy, base.Command):
  # pylint: disable=line-too-long
  """Deploys a service configuration for the given service name.

  This command is used to deploy a service configuration for a service
  to Google Service Management. As input, it takes one or more paths
  to service configurations that should be uploaded. These configuration
  files can be Proto Descriptors, Open API (Swagger) specifications,
  or Google Service Configuration files in JSON or YAML formats.

  If a service name is present in multiple configuration files (given
  in the `host` field in OpenAPI specifications or the `name` field in
  Google Service Configuration files), the first one will take precedence.

  When deploying a new service configuration to an already-existing
  service, some safety checks will be made comparing the new configuration
  to the active configuration. If any actionable advice is provided, it
  will be printed out to the log, and the deployment will be halted. It is
  recommended that these warnings be addressed before proceeding, but they
  can be overridden with the --force flag.

  This command will block until deployment is complete unless the
  `--async` flag is passed.

  ## EXAMPLES
  To deploy a single Open API service configuration, run:

    $ {command} ~/my_app/openapi.json

  To run the deployment asynchronously (non-blocking), run:

    $ {command} ~/my_app/openapi.json --async

  To deploy a service config with a Proto, run:

    $ {command} ~/my_app/service-config.yaml ~/my_app/service-protos.pb
  """
  # pylint: enable=line-too-long

  def CheckPushAdvisor(self, force=False):
    """Run the Push Advisor and return whether the command should abort.

    Args:
      force: bool, if True, this method will return False even if warnings are
        generated.

    Returns:
      True if the deployment should be aborted due to warnings, otherwise
      False if it's safe to continue.
    """
    log_func = log.warning if force else log.error
    num_advices = self.ShowConfigReport(
        self.service_name, self.service_config_id, log_func
    )
    if num_advices > 0:
      if force:
        log_func(('{0}\n').format(FORCE_ADVICE_STRING))
      else:
        log_func(('{0}\n').format(ADVICE_STRING))
        return True

    return False
